from typing import List
import numpy as np
from urdfenvs.urdf_common.generic_robot import ControlMode
from urdfenvs.urdf_common.holonomic_robot import HolonomicRobot
from urdfenvs.urdf_common.differential_drive_robot import DifferentialDriveRobot

class GenericDiffDriveRobot(DifferentialDriveRobot):

    def __init__(
            self,
            urdf: str,
            mode: ControlMode,
            actuated_wheels: List[str],
            castor_wheels: List[str],
            wheel_radius: float,
            wheel_distance: float,
            spawn_offset: np.ndarray = np.array([0.0, 0.0, 0.15]),
            spawn_rotation: float = 0.0,
            facing_direction: str = 'x',
            not_actuated_joints: List[str] = [],
    ):
        super().__init__(-1,
            urdf,
                         mode,
                         actuated_wheels,
                         castor_wheels,
                         wheel_radius,
                         wheel_distance,
                         spawn_offset=spawn_offset,
                         spawn_rotation=spawn_rotation,
                         facing_direction=facing_direction,
                         not_actuated_joints=not_actuated_joints)

    def set_joint_names(self):
        self._joint_names = []
        for joint in self._urdf_robot._actuated_joints:
            if joint.name in self._castor_wheels:
                continue
            if joint.name in self._actuated_wheels:
                continue
            if joint.name in self._not_actuated_joints:
                continue
            self._joint_names.append(joint.name)
        self._joint_names = self._actuated_wheels + self._joint_names

    def set_acceleration_limits(self):
        acc_limit = np.ones(self._n) * 10
        self._limit_acc_j[0, :] = -acc_limit[0 : self.n()]
        self._limit_acc_j[1, :] = acc_limit[0 : self.n()]

    def check_state(self, pos, vel):
        if not isinstance(pos, np.ndarray) or not pos.size == self.ns():
            center_position = (self._limit_pos_j[0] + self._limit_pos_j[1])/2
            pos = center_position
        if not isinstance(vel, np.ndarray) or not vel.size == self.n():
            vel = np.zeros(self.n())
        return pos, vel



"""Module for lidar sensor simulation."""
import numpy as np
import pybullet as p
import gymnasium as gym

from urdfenvs.sensors.sensor import Sensor
from urdfenvs.urdf_common.helpers import add_shape, extract_link_id


class Lidar(Sensor):
    """
    The Lidar sensor senses the distance toward the next object. A maximum
    sensing distance and the number of rays can be set. The Rays are evenly
    distributed in a circle.

    Attributes
    ----------

    _nb_rays: int
        Number of lidar rays spread over 2 pi.
    _ray_length: float
        Length of a single ray, maximum detection distance.
    _link_id: int
        Link of robot to which the lidar is connected.
    _thetas: list
        Angles for which rays are emitted.
    _rel_positions: np.ndarray
        Relative positions of first obstacle for each ray (x, y).
    _raw_data: bool
        Switch whether relative positions or raw distances are returned.
    _distance: np.ndarray
        Raw distance information for rays.
    """

    def __init__(
        self,
        link_name,
        nb_rays=10,
        ray_length=10.0,
        raw_data=True,
        angle_limits: np.ndarray = np.array([-np.pi, np.pi]),
        plotting_interval: int = -1,
        variance: float = 0.0,
    ):
        super().__init__(
            "LidarSensor",
            variance=variance,
            plotting_interval=plotting_interval,
        )
        self._nb_rays = nb_rays
        self._raw_data = raw_data
        self._ray_length = ray_length
        self._link_name = link_name
        self._link_id = None
        if isinstance(link_name, int):
            self._link_id = link_name
        self._angle_limits = angle_limits
        self._thetas = [
            angle_limits[0]
            + i * (angle_limits[1] - angle_limits[0]) / self._nb_rays
            for i in range(self._nb_rays)
        ]
        self._rel_positions = np.zeros(2 * nb_rays)
        self._distances = np.zeros(nb_rays)
        self._sphere_ids = [
            -1,
        ] * self._nb_rays

    def get_observation_size(self):
        """Getter for the dimension of the observation space."""
        if self._raw_data:
            return self._nb_rays
        return self._nb_rays * 2

    def get_observation_space(self, obstacles: dict, goals: dict):
        """Create observation space, all observations should be inside the
        observation space."""
        observation_space = gym.spaces.Box(
            -self._ray_length - 0.01,
            self._ray_length + 0.01,
            shape=(self.get_observation_size(),),
            dtype=float,
        )
        return gym.spaces.Dict({self._name: observation_space})

    def sense(self, robot, obstacles: dict, goals: dict, t: float):
        """Sense the distance toward the next object with the Lidar."""
        self._call_counter += 1
        if not self._link_id:
            self._link_id = extract_link_id(robot, self._link_name)
        link_state = p.getLinkState(robot, self._link_id)

        lidar_position = link_state[0]
        ray_start = lidar_position
        yaw = p.getEulerFromQuaternion(link_state[1])[2]
        for i, theta in enumerate(self._thetas):
            ray_end = np.array(ray_start) + self._ray_length * np.array(
                [np.cos(theta + yaw), np.sin(theta + yaw), 0.0]
            )
            lidar = p.rayTest(ray_start, ray_end)
            true_rel_positions = (
                lidar[0][2]
                * self._ray_length
                * np.array([np.cos(theta + yaw), np.sin(theta + yaw)])
            )
            noisy_rel_positions = np.random.normal(
                true_rel_positions, self._variance
            )

            self._rel_positions[2 * i : 2 * i + 2] = noisy_rel_positions
            self._distances[i] = np.linalg.norm(
                self._rel_positions[2 * i : 2 * i + 2]
            )
        if (
            self._plotting_interval > 0
            and self._call_counter % self._plotting_interval == 0
        ):
            self.update_lidar_spheres(lidar_position)
        if self._raw_data:
            return self._distances
        return self._rel_positions

    def init_lidar_spheres(self, lidar_position):
        """
        Initialize the Lidar spheres to visualize the sensing in bullet.
        The visual spheres are initialized once and then update in the
        update function. The relative positions are augmented by a third
        column to have a full position, including the z-coordinate.

        Parameters
        ------------
        lidar_position : np.ndarray
            The position of the lidar sensor link.
        """
        q = lidar_position
        q_obs = self._rel_positions.reshape(self._nb_rays, 2)
        q_obs = np.append(q_obs, np.zeros((self._nb_rays, 1)), axis=1)
        for ray_id in range(self._nb_rays):
            body_id_sphere = add_shape(
                "sphere",
                size=[0.05],
                color=[0.0, 0.0, 0.0, 0.8],
                position=q + q_obs[ray_id],
                with_collision_shape=False,
            )
            self._sphere_ids[ray_id] = body_id_sphere

    def update_lidar_spheres(self, lidar_position):
        """
        Updates the position of the spheres visualizing the sensing with lidar.
        If the spheres have not been initialized, the init_lidar_spheres
        function is called.

        Parameters
        ------------
        lidar_position : np.ndarray
            The position of the lidar sensor link.
        """
        if self._sphere_ids[0] == -1:
            self.init_lidar_spheres(lidar_position)
        q = lidar_position
        # Reshape and add z-values to the sensor data.
        q_obs = self._rel_positions.reshape(self._nb_rays, 2)
        q_obs = np.append(q_obs, np.zeros((self._nb_rays, 1)), axis=1)
        for ray_id in range(self._nb_rays):
            p.resetBasePositionAndOrientation(
                int(self._sphere_ids[ray_id]), q + q_obs[ray_id], [0, 0, 0, 1]
            )



import gymnasium as gym
import numpy as np

class SB3Float32ActionWrapper(gym.Wrapper):
    """
    A wrapper to convert action space to float32.

    Unfortunately, Stable Baselines3 has a problem with `float64` 
    action spaces in DDPG, TD3 and SAC algorithms. This wrapper 
    solves that problem by converting the `dType` of action space 
    to `float32`. Use it like `env = SB3Float32ActionWrapper(env)`.

    """
    def __init__(self, env):
        super().__init__(env)
        self.action_space = gym.spaces.Box(
            low=self.env.action_space.low.astype(np.float32),
            high=self.env.action_space.high.astype(np.float32),
            dtype=np.float32,
        )

    def step(self, action):
        action = action.astype(np.float32)
        obs, reward, done, truncated, info = self.env.step(action)
        return obs, reward, done, truncated, info


import pybullet as p
import gymnasium as gym
import numpy as np

from urdfenvs.urdf_common.generic_robot import GenericRobot


class HolonomicRobot(GenericRobot):
    """Generic holonomic robot."""

    def reset(
            self,
            pos: np.ndarray,
            vel: np.ndarray,
            mount_position: np.ndarray,
            mount_orientation: np.ndarray,) -> None:

        if hasattr(self, "_robot"):
            p.removeBody(self._robot)
        self._robot = p.loadURDF(
            fileName=self._urdf_file,
            basePosition=mount_position.tolist(),
            baseOrientation=mount_orientation.tolist(),
            flags=p.URDF_USE_SELF_COLLISION_EXCLUDE_PARENT,
            useFixedBase=True,
        )
        self.set_joint_names()
        self.extract_joint_ids()
        for i in range(self._n):
            p.resetJointState(
                self._robot,
                self._robot_joints[i],
                pos[i],
                targetVelocity=vel[i],
            )
        self.update_state()
        self._integrated_velocities = vel

    def read_limits(self) -> None:
        """
        Set position, velocity, acceleration and
        motor torque lower en upper limits
        """
        self._limit_pos_j = np.zeros((2, self._n))
        self._limit_vel_j = np.zeros((2, self._n))
        self._limit_tor_j = np.zeros((2, self._n))
        self._limit_acc_j = np.zeros((2, self._n))
        for i, j in enumerate(self._urdf_joints):
            joint = self._urdf_robot.robot.joints[j]
            self._limit_pos_j[0, i] = joint.limit.lower
            self._limit_pos_j[1, i] = joint.limit.upper
            self._limit_vel_j[0, i] = -joint.limit.velocity
            self._limit_vel_j[1, i] = joint.limit.velocity
            self._limit_tor_j[0, i] = -joint.limit.effort
            self._limit_tor_j[1, i] = joint.limit.effort
        self.set_acceleration_limits()

    def get_observation_space(self) -> gym.spaces.Dict:
        """
        Gets the observation space for a holonomic robot.

        The observation space is represented as a dictionary.
        `joint_state` containing:
        `position` the concatenated positions of joints in
        their local configuration space.
        `velocity` the concatenated velocities of joints in
        their local configuration space.
        """
        return gym.spaces.Dict(
            {
                "joint_state": gym.spaces.Dict({
                    "position": gym.spaces.Box(
                        low=self._limit_pos_j[0, :],
                        high=self._limit_pos_j[1, :],
                        dtype=float,
                    ),
                    "velocity": gym.spaces.Box(
                        low=self._limit_vel_j[0, :],
                        high=self._limit_vel_j[1, :],
                        dtype=float,
                    ),

                }),
            }
        )

    def apply_torque_action(self, torques: np.ndarray) -> None:
        for i in range(self._n):
            p.setJointMotorControl2(
                self._robot,
                self._robot_joints[i],
                controlMode=p.TORQUE_CONTROL,
                force=torques[i],
            )

    def apply_velocity_action(self, vels: np.ndarray) -> None:
        for i in range(self._n):
            p.setJointMotorControl2(
                self._robot,
                self._robot_joints[i],
                controlMode=p.VELOCITY_CONTROL,
                targetVelocity=vels[i],
            )

    def apply_acceleration_action(self, accs: np.ndarray, dt: float) -> None:
        self._integrated_velocities += dt * accs
        self.apply_velocity_action(self._integrated_velocities)

    def update_state(self) -> None:
        """
        Updates the robot joint_state.

        The robot joint_state is stored in the dictionary self.state,
        which contains:
       `position`: np.array([joint_position_0, ..., joint_position_n-1)
           the joints 0 to n-1 have al 1-dimensional configuration space
           joint_position_i = (position in local configuration space)
       `velocity`: np.array([joint_velocity_0, ..., joint_velocity_n-1])
           the joints 0 to n-1 have al one dimensional configuration space
           joint_velocity_i = (position in local configuration space)
       """

        # Get Joint Configurations
        joint_pos_list = []
        joint_vel_list = []
        for i in range(self._n):
            pos, vel, _, _ = p.getJointState(self._robot, self._robot_joints[i])
            joint_pos_list.append(pos)
            joint_vel_list.append(vel)
        joint_pos = np.array(joint_pos_list)
        joint_vel = np.array(joint_vel_list)

        # Concatenate position, orientation, velocity
        self.state = {"joint_state": {"position": joint_pos,
                      "velocity": joint_vel}}



from mpscenes.goals.static_sub_goal import StaticSubGoal
from mpscenes.goals.dynamic_sub_goal import DynamicSubGoal
from mpscenes.goals.goal_composition import GoalComposition

goal1Dict = {
    "weight": 1.0,
    "is_primary_goal": True,
    "indices": [0, 1, 2],
    "parent_link": 0,
    "child_link": 3,
    "desired_position": [1, 0, 0.1],
    "epsilon": 0.2,
    "type": "staticSubGoal",
    'high': [3, 3, 0.1],
    'low': [-3, -3, 0.1],
}

goal1 = StaticSubGoal(name="goal1", content_dict=goal1Dict)
dynamicGoalDict = {
    "weight": 1.0,
    "is_primary_goal": True,
    "indices": [0, 1, 2],
    "parent_link": 0,
    "child_link": 3,
    "trajectory": ["0.5", "0.2 + 0.2 * sp.sin(0.3 * t)", "0.4"],
    "epsilon": 0.08,
    "type": "analyticSubGoal",
}
dynamicGoal = DynamicSubGoal(name="goal2", content_dict=dynamicGoalDict)
splineDict = {
    "degree": 2,
    "controlPoints": [[0.0, -0.0, 0.2], [3.0, 0.0, 2.2], [3.0, 3.0, 1.2]],
    "duration": 10,
}
splineGoalDict = {
    "weight": 1.0,
    "is_primary_goal": True,
    "indices": [0, 1, 2],
    "parent_link": 0,
    "child_link": 3,
    "trajectory": splineDict,
    "epsilon": 0.08,
    "type": "splineSubGoal",
}
splineGoal = DynamicSubGoal(name="goal3", content_dict=splineGoalDict)
goal_composition_dict = {
    'subGoal0': {
        "weight": 1.0,
        "is_primary_goal": True,
        "indices": [0, 1, 2],
        "parent_link": 0,
        "child_link": 3,
        "desired_position": [1, 0, 0.1],
        "epsilon": 0.02,
        "type": "staticSubGoal",
    }
}
goal_composition = GoalComposition(
    name="goal_composition",
    content_dict=goal_composition_dict
)



"""Module for occupancy sensor simulation."""
from time import perf_counter
import logging

import numpy as np
import pybullet
import gymnasium as gym

from urdfenvs.sensors.grid_sensor import GridSensor


class OccupancySensor(GridSensor):
    def __init__(
        self,
        limits: np.ndarray = np.array([[-1, -1], [-1, 1], [-1, 1]]),
        resolution: np.ndarray = np.array([10, 10, 10], dtype=int),
        interval: int = -1,
        variance: float = 0.0,
        plotting_interval: int = -1,
    ):
        super().__init__(
            limits=limits,
            resolution=resolution,
            interval=interval,
            name="Occupancy",
            variance=variance,
            plotting_interval=plotting_interval,
        )
        self._voxel_ids = [
            -1,
        ] * self.number_of_voxels()
        self._voxel_size = self.voxel_size() * 1.0
        self._bullet_ids = []

    def get_observation_space(self, obstacles: dict, goals: dict):
        """Create observation space, all observations should be inside the
        observation space."""
        observation_space = gym.spaces.Box(
            0,
            1,
            shape=self.get_observation_size(),
            dtype=int,
        )
        return gym.spaces.Dict({self._name: observation_space})

    def sense(self, robot, obstacles: dict, goals: dict, t: float):
        self._compute_call_counter += 1
        self._call_counter += 1
        if not (
            self._computed
            and (self._interval < 0 or self._call_counter % self._interval != 0)
        ):
            start_time = perf_counter()
            distances = self.distances(obstacles, t)
            self._grid_values = np.array(distances <= 0.0, dtype=int).reshape(
                self._resolution
            )
            end_time = perf_counter()

            logging.info(f"Computed Occupancy in {end_time-start_time} s")
            self._computed = True
        if (
            self._plotting_interval > 0
            and self._call_counter % self._plotting_interval == 0
        ):
            self.update_occupancy_visualization()
        return self._grid_values

    def update_occupancy_visualization(self):
        """
        Updates the position of the boxes visualizing the occupancy.
        If the boxes have not been initialized, the function
        init_occupancy_visualization(self) is called.

        Parameters
        ------------
        """
        for bullet_id in self._bullet_ids:
            pybullet.removeBody(bullet_id)
        self._bullet_ids = []
        grid_values_flat = self._grid_values.reshape((-1, 1))
        voxel_positions = []
        for voxel_id in range(self.number_of_voxels()):
            if grid_values_flat[voxel_id] == 1:
                voxel_positions.append(self._mesh_flat[voxel_id].tolist())

        nb_occupied_cells = len(voxel_positions)
        for i in range(0, nb_occupied_cells, 16):
            voxel_positions_chunk = voxel_positions[
                i : min(i + 16, nb_occupied_cells)
            ]
            half_extens = np.tile(
                self._voxel_size * 0.5, (nb_occupied_cells, 1)
            ).tolist()
            shape_types = [pybullet.GEOM_BOX] * len(voxel_positions_chunk)
            half_extens = np.tile(
                self._voxel_size * 0.5, (len(voxel_positions_chunk), 1)
            ).tolist()
            visual_shape_id = pybullet.createVisualShapeArray(
                shape_types,
                halfExtents=half_extens,
                visualFramePositions=voxel_positions_chunk,
            )
            bullet_id = pybullet.createMultiBody(
                baseMass=0,
                baseCollisionShapeIndex=-1,
                baseVisualShapeIndex=visual_shape_id,
                useMaximalCoordinates=False,
            )
            pybullet.changeVisualShape(
                bullet_id, -1, rgbaColor=[0.0, 0.0, 0.0, 0.3]
            )
            self._bullet_ids.append(bullet_id)



from mpscenes.obstacles.sphere_obstacle import SphereObstacle
from mpscenes.obstacles.dynamic_sphere_obstacle import DynamicSphereObstacle
from mpscenes.obstacles.urdf_obstacle import UrdfObstacle
from mpscenes.obstacles.box_obstacle import BoxObstacle
from mpscenes.obstacles.cylinder_obstacle import CylinderObstacle

import os

obst1Dict = {
    "type": "sphere",
    "geometry": {"position": [2.0, 2.0, 1.0], "radius": 1.0},
    "rgba": [0.3, 0.5, 0.6, 1.0],
}
sphereObst1 = SphereObstacle(name="simpleSphere", content_dict=obst1Dict)
obst2Dict = {
    "type": "sphere",
    "movable": False,
    "geometry": {"position": [2.0, -0.0, 0.0], "radius": 0.2},
}
sphereObst2 = SphereObstacle(name="simpleSphere", content_dict=obst2Dict)
urdfObst1Dict = {
    "type": "urdf",
    "geometry": {"position": [1.5, 0.0, 0.05]},
    "urdf": os.path.join(os.path.dirname(__file__), "obstacle_data/duck.urdf"),
}
urdfObst1 = UrdfObstacle(name="duckUrdf", content_dict=urdfObst1Dict)
dynamicObst1Dict = {
    "type": "sphere",
    "geometry": {"trajectory": ["2.0 - 0.1 * t", "-0.0", "0.1"], "radius": 0.2},
    "movable": False,
}
dynamicSphereObst1 = DynamicSphereObstacle(
    name="simpleSphere", content_dict=dynamicObst1Dict
)
dynamicObst2Dict = {
    "type": "analyticSphere",
    "geometry": {"trajectory": ["0.6", "0.5 - 0.1 * t", "0.8"], "radius": 0.2},
    "movable": False,
}
dynamicSphereObst2 = DynamicSphereObstacle(
    name="simpleSphere", content_dict=dynamicObst2Dict
)
splineDict = {
    "degree": 2,
    "controlPoints": [[0.0, 1.0, 1.0], [1.0, 1.0, 1.0], [1.0, 1.0, 0.0]],
    "duration": 10,
}
dynamicObst3Dict = {
    "type": "splineSphere",
    "geometry": {"trajectory": splineDict, "radius": 0.2},
}
dynamicSphereObst3 = DynamicSphereObstacle(
    name="simpleSphere", content_dict=dynamicObst3Dict
)
movable_obstacle_dict = {
    'type': 'box',
    'geometry': {
        'position' : [2.0, 0.0, 2.0],
        'width': 0.2,
        'height': 0.2,
        'length': 0.2,
    },
    'movable': True,
    'high': {
         'position' : [5.0, 5.0, 1.0],
        'width': 0.2,
        'height': 0.2,
        'length': 0.2,
    },
    'low': {
        'position' : [0.0, 0.0, 0.5],
        'width': 0.2,
        'height': 0.2,
        'length': 0.2,
    }
}
movable_obstacle = BoxObstacle(name="movable_box", content_dict=movable_obstacle_dict)


wall_length = 10
wall_obstacles_dicts = [
    {
        'type': 'box', 
         'geometry': {
             'position': [wall_length/2.0, 0.0, 0.4], 'width': wall_length, 'height': 0.8, 'length': 0.1
        },
        'high': {
            'position' : [wall_length/2.0, 0.0, 0.4],
            'width': wall_length,
            'height': 0.8,
            'length': 0.1,
        },
        'low': {
            'position' : [wall_length/2.0, 0.0, 0.4],
            'width': wall_length,
            'height': 0.8,
            'length': 0.1,
        },
    },
    {
        'type': 'box', 
         'geometry': {
             'position': [0.0, wall_length/2.0, 0.4], 'width': 0.1, 'height': 0.8, 'length': wall_length
        },
        'high': {
            'position' : [0.0, wall_length/2.0, 0.4],
            'width': 0.1,
            'height': 0.8,
            'length': wall_length,
        },
        'low': {
            'position' : [0.0, wall_length/2.0, 0.4],
            'width': 0.1,
            'height': 0.8,
            'length': wall_length,
        },
    },
    {
        'type': 'box', 
         'geometry': {
             'position': [0.0, -wall_length/2.0, 0.4], 'width': 0.1, 'height': 0.8, 'length': wall_length
        },
        'high': {
            'position' : [0.0, -wall_length/2.0, 0.4],
            'width': 0.1,
            'height': 0.8,
            'length': wall_length,
        },
        'low': {
            'position' : [0.0, -wall_length/2.0, 0.4],
            'width': 0.1,
            'height': 0.8,
            'length': wall_length,
        },
    },
    {
        'type': 'box', 
         'geometry': {
             'position': [-wall_length/2.0, 0.0, 0.4], 'width': wall_length, 'height': 0.8, 'length': 0.1
        },
        'high': {
            'position' : [-wall_length/2.0, 0.0, 0.4],
            'width': wall_length,
            'height': 0.8,
            'length': 0.1,
        },
        'low': {
            'position' : [-wall_length/2.0, 0.0, 0.4],
            'width': wall_length,
            'height': 0.8,
            'length': 0.1,
        },
    },
]

wall_obstacles = [BoxObstacle(name=f"wall_{i}", content_dict=obst_dict) for i, obst_dict in enumerate(wall_obstacles_dicts)]

cylinder_obstacle_dict = {
    "type": "cylinder",
    "movable": False,
    "geometry": {
        "position": [2.0, -3.0, 0.0],
        "radius": 0.5,
        "height": 2.0,
    },
    "rgba": [0.1, 0.3, 0.3, 1.0],
}
cylinder_obstacle = CylinderObstacle(
    name="cylinder_obstacle",
    content_dict=cylinder_obstacle_dict
)



"""Module for signed distance field sensor simulation."""
from time import perf_counter
import logging
import numpy as np
import gymnasium as gym

from urdfenvs.sensors.grid_sensor import GridSensor


class LinkIdNotFoundError(Exception):
    pass


class SDFSensor(GridSensor):
    def __init__(
        self,
        limits: np.ndarray = np.array([[-1, -1], [-1, 1], [-1, 1]]),
        resolution: np.ndarray = np.array([10, 10, 10], dtype=int),
        interval: int = -1,
        variance: float = 0.0,
    ):
        super().__init__(
            limits=limits,
            resolution=resolution,
            interval=interval,
            name="SDFSensor",
            variance=variance,
        )

    def get_observation_space(self, obstacles: dict, goals: dict):
        """Create observation space, all observations should be inside the
        observation space."""
        observation_space = gym.spaces.Box(
            0.0,
            np.inf,
            shape=self.get_observation_size(),
            dtype=float,
        )
        return gym.spaces.Dict({self._name: observation_space})

    def sense(self, robot, obstacles: dict, goals: dict, t: float):
        self._call_counter += 1
        if self._computed and (
            self._interval < 0 or self._call_counter % self._interval != 0
        ):
            return self._grid_values
        start_time = perf_counter()
        distances = self.distances(obstacles, t)
        self._grid_values = np.maximum(distances, 0.0).reshape(self._resolution)
        end_time = perf_counter()

        logging.info(f"Computed SDF in {end_time-start_time} s")
        self._computed = True
        return self._grid_values



import numpy as np
from urdfenvs.urdf_common.holonomic_robot import HolonomicRobot
import os


class GenericUrdfReacher(HolonomicRobot):
    def __init__(self, urdf, mode):
        super().__init__(-1, urdf, mode=mode)

    def set_joint_names(self):
        # TODO Replace it with a automated extraction
        self._joint_names = [joint.name for joint in self._urdf_robot._actuated_joints]

    def set_acceleration_limits(self):
        acc_limit = np.array(
            [1.0, 1.0, 15.0, 15.0, 7.5, 10.0, 12.5, 15.0, 20.0, 20.0, 1.0, 1.0]
        )
        self._limit_acc_j[0, :] = -acc_limit[0 : self.n()]
        self._limit_acc_j[1, :] = acc_limit[0 : self.n()]

    def check_state(self, pos, vel):
        if not isinstance(pos, np.ndarray) or not pos.size == self.n():
            center_position = (self._limit_pos_j[0] + self._limit_pos_j[1])/2
            pos = center_position
        if not isinstance(vel, np.ndarray) or not vel.size == self.n():
            vel = np.zeros(self.n())
        return pos, vel



import time
from copy import deepcopy
import warnings
import logging
from typing import List, Type, Union, Optional, Tuple

import dill
import pybullet as p
import numpy as np
import gymnasium as gym

from mpscenes.obstacles.collision_obstacle import CollisionObstacle
from mpscenes.goals.goal_composition import GoalComposition
from mpscenes.goals.sub_goal import SubGoal

from urdfenvs.urdf_common.plane import Plane
from urdfenvs.sensors.sensor import Sensor
from urdfenvs.urdf_common.generic_robot import GenericRobot
from urdfenvs.urdf_common.reward import Reward
from urdfenvs.urdf_common.helpers import add_shape, get_transformation_matrix, matrix_to_quaternion


class WrongObservationError(Exception):
    pass


class WrongActionError(Exception):
    pass


def check_observation(obs, ob):
    for key, value in ob.items():
        if isinstance(value, dict):
            check_observation(obs[key], value)
        elif isinstance(value, np.ndarray):
            if isinstance(obs[key], gym.spaces.Discrete):
                continue
            if not obs[key].contains(value):
                s = f"key: {key}: {value} not in {obs[key]}"
                if np.any(value < obs[key].low):
                    index = np.where(value < obs[key].low)[0]
                    value_at_index = value[index]
                    s += f"\nAt index {index.tolist()}: {value_at_index} < {obs[key].low[index]}"
                if np.any(value > obs[key].high):
                    index = np.where(value > obs[key].high)[0]
                    value_at_index = value[index]
                    s += f"\nAt index {index.tolist()}: {value_at_index} > {obs[key].high[index]}"

                raise WrongObservationError(s)
        else:
            raise Exception(f"Observation checking failed for key:{key} value:{value}.")


class UrdfEnv(gym.Env):
    """Generic urdf-environment for OpenAI-Gym"""

    def __init__(
        self,
        robots: List[GenericRobot],
        render: bool = False,
        enforce_real_time: Optional[bool] = None,
        dt: float = 0.01,
        observation_checking=True,
    ) -> None:
        """Constructor for environment.

        Variables are set and the pyhsics engine is initiated. Either with
        rendering (p.GUI) or without (p.DIRECT). Note that rendering slows
        down the simulation.

        Parameters:

        robot: Robot instance to be simulated
        render: Flag if simulator should render
        dt: Time step for pyhsics engine
        """
        assert len(robots) > 0
        self._dt: float = dt
        self._t: float = 0.0
        self._robots: List[GenericRobot] = robots
        self._render: bool = render
        self._enforce_real_time: bool = render if enforce_real_time is None else enforce_real_time
        self._done: bool = False
        self._info: dict = {}
        self._num_sub_steps: float = 20
        self._obsts: dict = {}
        self._collision_links: dict = {}
        self._collision_links_poses: dict = {}
        self._goals: dict = {}
        self._space_set = False
        self._observation_checking = observation_checking
        self._reward_calculator = None
        self.sensors = (
            []
        )
        self.connect_physics_engine()
        self._obsts = {}
        self._collision_links = {}
        self._goals = {}
        self.set_spaces()

    def connect_physics_engine(self):
        if self._render:
            self._cid = p.connect(p.GUI)
            p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0)
        else:
            self._cid = p.connect(p.DIRECT)
        p.setPhysicsEngineParameter(
            fixedTimeStep=self._dt, numSubSteps=self._num_sub_steps
        )
        self.plane = Plane()
        p.setGravity(0, 0, -10.0)

    def load_environment(self):
        old_obsts = deepcopy(list(self._obsts.values()))
        self._obsts = {}
        for obst in old_obsts:
            self.add_obstacle(obst)
        old_goals = deepcopy(list(self._goals.values()))
        self._goals = {}
        for goal in old_goals:
            self.add_goal(goal)
        #TODO load collision links

    @classmethod
    def load(cls: Type["UrdfEnv"], file_name: str, render: bool = False) -> "UrdfEnv":
        with open(file_name, "rb") as f:
            env: "UrdfEnv" =  dill.load(f)
        env._render = render
        env.connect_physics_engine()
        env.reset()
        env.load_environment()
        return env

    def set_reward_calculator(self, reward_calculator: Reward) -> None:
        self._reward_calculator = reward_calculator

    def n(self) -> int:
        return sum(self.n_per_robot())

    def n_per_robot(self) -> list:
        return [robot.n() for robot in self._robots]

    def ns_per_robot(self) -> list:
        return [robot.ns() for robot in self._robots]

    def dt(self) -> float:
        return self._dt

    def t(self) -> float:
        return self._t

    def get_camera_configuration(self) -> tuple:
        full_camera_configuration = p.getDebugVisualizerCamera()
        camera_yaw = full_camera_configuration[8]
        camera_pitch = full_camera_configuration[9]
        camera_distance = full_camera_configuration[10]
        camera_target_position = full_camera_configuration[11]
        return (
            camera_distance,
            camera_yaw,
            camera_pitch,
            camera_target_position,
        )

    def reconfigure_camera(
        self,
        camera_distance: float,
        camera_yaw: float,
        camera_pitch: float,
        camera_target_position: tuple,
    ) -> None:
        p.resetDebugVisualizerCamera(
            cameraDistance=camera_distance,
            cameraYaw=camera_yaw,
            cameraPitch=camera_pitch,
            cameraTargetPosition=camera_target_position,
        )

    def start_video_recording(self, file_name: str) -> None:
        if self._render:
            p.startStateLogging(p.STATE_LOGGING_VIDEO_MP4, file_name)
        else:
            logging.warning("Video recording requires rendering to be active.")

    def stop_video_recording(self) -> None:
        if self._render:
            p.stopStateLogging()

    def set_spaces(self) -> None:
        """Set observation and action space."""
        observation_space_as_dict = {}
        action_space_as_dict = {}

        for i, robot in enumerate(self._robots):
            (obs_space_robot_i, action_space_robot_i) = robot.get_spaces()
            obs_space_robot_i = dict(obs_space_robot_i)
            for sensor in robot._sensors:

                self.sensors.append(sensor)  # Add the sensor to the list of sensors.

                obs_space_robot_i.update(
                    sensor.get_observation_space(self._obsts, self._goals)
                )
            observation_space_as_dict[f"robot_{i}"] = gym.spaces.Dict(obs_space_robot_i)
            action_space_as_dict[f"robot_{i}"] = action_space_robot_i

        self.observation_space = gym.spaces.Dict(observation_space_as_dict)
        action_space = gym.spaces.Dict(action_space_as_dict)
        self.action_space = gym.spaces.flatten_space(action_space)

    def step(self, action: np.ndarray):
        step_start = time.perf_counter()
        self._t += self.dt()
        # Feed action to the robot and get observation of robot's state

        if not self.action_space.contains(action):
            self._done = True
            self._info = {"action_limits": f"{action} not in {self.action_space}"}

        action_id = 0
        for robot in self._robots:
            action_robot = action[action_id : action_id + robot.n()]
            robot.apply_action(action_robot, self.dt())
            action_id += robot.n()

        self.update_obstacles()
        self.update_goals()
        self.update_collision_links()
        p.stepSimulation(self._cid)
        for robot_id, robot in enumerate(self._robots):
            contacts = p.getContactPoints(robot._robot)
            for contact_info in contacts:
                body_b = contact_info[2]
                if body_b in self._obsts:
                    message = f"Collision occured at {round(self.t(), 2)} " \
                        f"between robot {robot_id} and obstacle " \
                        f"with id {body_b}"
                    self._info = {"Collision": message}
                    self._done = True
                    break
        ob = self._get_ob()

        # Calculate the reward.
        # If there is no reward object, then the reward is 1.0.
        if self._reward_calculator is not None:
            reward = self._reward_calculator.calculate_reward(ob)
        else:
            reward = 1.0

        if self._render:
            self.render()

        terminated = self._done
        truncated = self._get_truncated()
        step_end = time.perf_counter()
        step_time = step_end - step_start
        if self._enforce_real_time:
            sleep_time = max(0.0, self.dt() - step_time)
            time.sleep(sleep_time)
        step_final_end = time.perf_counter()
        total_step_time = step_final_end - step_start
        real_time_factor = self.dt()/total_step_time
        logging.info(f"Real time factor {real_time_factor}")
        return ob, reward, terminated, truncated, self._info
    
    def _get_truncated(self) -> bool:
        # TODO: Implement Truncated. 
        return False
    
    def _get_ob(self) -> dict:
        """Compose the observation."""
        observation = {}
        for i, robot in enumerate(self._robots):
            obs = robot.get_observation(self._obsts, self._goals, self.t())

            observation[f"robot_{i}"] = obs
        if hasattr(self, "observation_space"):
            if (
                not self.observation_space.contains(observation)
                and self._observation_checking
            ):
                try:
                    check_observation(self.observation_space, observation)
                except WrongObservationError as e:
                    self._done = True
                    self._info = {"observation_limits": str(e)}
        return observation

    def shuffle_obstacles(self) -> dict:
        obstacle_dict = {}
        for obst_id, obst in self._obsts.items():
            obst.shuffle()
            obstacle_dict[obst.name()] = obst.dict()
        self.update_obstacles()
        return obstacle_dict

    def shuffle_goals(self) -> dict:
        goal_dict = {}
        for goal_id, goal in self._goals.items():
            goal.shuffle()
            goal_dict[goal.name()] = goal.dict()
        self.update_goals()
        return goal_dict

    def empty_scene(self) -> None:
        for goal_id in self._goals:
            p.removeBody(goal_id)
        self._goals = {}
        for obst_id in self._obsts:
            p.removeBody(obst_id)
        self._obsts = {}

    def update_obstacles(self):
        for obst_id, obst in self._obsts.items():
            if obst.movable():
                continue
            try:
                pos = obst.position(t=self.t()).tolist()
                vel = obst.velocity(t=self.t()).tolist()
                ori = [0, 0, 0, 1]
                p.resetBasePositionAndOrientation(obst_id, pos, ori)
                p.resetBaseVelocity(obst_id, linearVelocity=vel)
            except Exception:
                continue

    def update_collision_links(self) -> None:
        for visual_shape_id, info in self._collision_links.items():
            link_state = p.getLinkState(info[0], info[1])
            link_position = link_state[0]
            link_ori = np.array(link_state[1])
            transformation_matrix = get_transformation_matrix(link_ori, link_position)
            total_transformation = np.dot(transformation_matrix, info[2])
            self._collision_links_poses[f"{info[3]}_{info[1]}_{info[4]}"] = total_transformation
            translation, rotation = matrix_to_quaternion(total_transformation, ordering='xyzw')
            p.resetBasePositionAndOrientation(
                visual_shape_id, translation, rotation
            )

    def collision_links_poses(self, position_only: bool=False) -> dict:
        if position_only:
            result_dict = {}
            for key, value in self._collision_links_poses.items():
                if value is None:
                    result_dict[key] = None
                elif isinstance(value, np.ndarray):
                    result_dict[key] = value[0:3, 3]
            return result_dict
        return self._collision_links_poses

    def update_goals(self):
        for goal_id, goal in self._goals.items():
            try:
                pos = goal.position(t=self.t()).tolist()
                vel = goal.velocity(t=self.t()).tolist()
                ori = [0, 0, 0, 1]
                p.resetBasePositionAndOrientation(goal_id, pos, ori)
                p.resetBaseVelocity(goal_id, linearVelocity=vel)
            except Exception:
                continue

    def add_obstacle(self, obst: CollisionObstacle) -> None:
        """Adds obstacle to the simulation environment.

        Parameters
        ----------

        obst: Obstacle from mpscenes
        """
        # add obstacle to environment
        if obst.type() == "urdf":
            obst_id = add_shape(obst.type(), obst.size(), urdf=obst.urdf())
        else:
            obst_id = add_shape(
                obst.type(),
                obst.size(),
                obst.rgba().tolist(),
                position=obst.position(),
                movable=obst.movable(),
            )
        self._obsts[obst_id] = obst
        if self._t != 0.0:
            warnings.warn("Adding an object while the simulation already started")

    def reset_obstacles(self) -> None:
        for obst_id, obstacle in self._obsts.items():
            if obstacle.type() == "urdf":
                pos = obstacle.position()
                vel = obstacle.position()
            else:
                pos = obstacle.position(t=0).tolist()
                vel = obstacle.velocity(t=0).tolist()
            ori = [0, 0, 0, 1]
            p.resetBasePositionAndOrientation(obst_id, pos, ori)
            p.resetBaseVelocity(obst_id, linearVelocity=vel)

    def reset_goals(self) -> None:
        for goal_id, goal in self._goals.items():
            pos = goal.position(t=0).tolist()
            vel = goal.velocity(t=0).tolist()
            ori = [0, 0, 0, 1]
            p.resetBasePositionAndOrientation(goal_id, pos, ori)
            p.resetBaseVelocity(goal_id, linearVelocity=vel)

    def get_obstacles(self) -> dict:
        return self._obsts

    def add_collision_link(
        self,
        robot_index: int = 0,
        link_index: Union[int, str] = 0,
        sphere_on_link_index: int=0,
        shape_type: str = "sphere",
        size: Optional[List[float]] = None,
        link_transformation: Optional[np.ndarray] = None,
    ) -> int:
        if size is None:
            size = [1.0]
        if link_transformation is None:
            link_transformation = np.identity(4)
        rgba_color = [1.0, 1.0, 0.0, 0.3]
        bullet_id = add_shape(
                shape_type,
                size,
                rgba_color,
                with_collision_shape=False,
            )
        if isinstance(link_index, str):
            link_index = self._robots[robot_index]._link_names.index(link_index)
        self._collision_links[bullet_id] = (
            self._robots[robot_index]._robot,
            link_index,
            link_transformation,
        )
        self._collision_links_poses[f"{robot_index}_{link_index}_{sphere_on_link_index}"] = None
        self._collision_links[bullet_id] = (
            self._robots[robot_index]._robot,
            link_index,
            link_transformation,
            robot_index,
            sphere_on_link_index,
        )
        return bullet_id

    def add_debug_shape(
        self,
        position: Tuple[float],
        orientation: Tuple[float],
        shape_type: str = "sphere",
        size: Optional[List[float]] = None,
        rgba_color : Optional[List[float]] = None
    ) -> None:
        if size is None:
            size = [1.0]
        add_shape(
                shape_type,
                size,
                rgba_color,
                with_collision_shape=False,
                orientation=orientation,
                position=position,
            )


    def add_sub_goal(self, goal: SubGoal) -> int:
        rgba_color = [0.0, 1.0, 0.0, 0.3]
        visual_shape_id = p.createVisualShape(
            p.GEOM_SPHERE, rgbaColor=rgba_color, radius=goal.epsilon()
        )
        collision_shape = -1
        base_position = [
            0,
        ] * 3
        for index in range(3):
            if index in goal.indices():
                base_position[index] = goal.position()[index]

        base_orientation = [0, 0, 0, 1]

        assert isinstance(base_position, list)
        assert isinstance(base_orientation, list)
        bullet_id = p.createMultiBody(
            0,
            collision_shape,
            visual_shape_id,
            base_position,
            base_orientation,
        )
        return bullet_id

    def add_goal(self, goal: Union[GoalComposition, SubGoal]) -> None:
        """Adds goal to the simulation environment.

        Parameters
        ----------

        goal: Goal from mpscenes
        """
        if isinstance(goal, GoalComposition):
            for sub_goal in goal.sub_goals():
                goal_id = self.add_sub_goal(sub_goal)
                self._goals[goal_id] = goal
        else:
            goal_id = self.add_sub_goal(goal)
            self._goals[goal_id] = goal


    def add_sensor(self, sensor: Sensor, robot_ids: List) -> None:
        """Adds sensor to the robot.

        Adding a sensor to the list of sensors.
        """
        for i in robot_ids:
            self._robots[i].add_sensor(sensor)

    def reset(
        self,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
        pos: Optional[np.ndarray] = None,
        vel: Optional[np.ndarray] = None,
        mount_positions: Optional[np.ndarray] = None,
        mount_orientations: Optional[np.ndarray] = None,
    ) -> tuple:
        """Resets the simulation and the robot.

        Parameters
        ----------

        pos: np.ndarray:
            Initial joint positions of the robots
        vel: np.ndarray:
            Initial joint velocities of the robots
        mount_position: np.ndarray:
            Mounting position for the robots
            This is ignored for mobile robots
        mount_orientation: np.ndarray:
            Mounting position for the robots
            This is ignored for mobile robots
        """
        super().reset(seed=seed, options=options)
        self._t = 0.0
        if mount_positions is None:
            mount_positions = np.tile(np.zeros(3), (len(self._robots), 1))
        self.mount_positions = mount_positions
        if mount_orientations is None:
            mount_orientations = np.tile(
                np.array([0.0, 0.0, 0.0, 1.0]), (len(self._robots), 1)
            )
        if pos is None:
            pos = np.tile(None, len(self._robots))
        if vel is None:
            vel = np.tile(None, len(self._robots))
        if len(pos.shape) == 1 and len(self._robots) == 1:
            pos = np.tile(pos, (1, 1))
        if len(vel.shape) == 1 and len(self._robots) == 1:
            vel = np.tile(vel, (1, 1))
        for i, robot in enumerate(self._robots):
            checked_position, checked_velocity = robot.check_state(pos[i], vel[i])
            robot.reset(
                pos=checked_position,
                vel=checked_velocity,
                mount_position=mount_positions[i],
                mount_orientation=mount_orientations[i],
            )
        self.reset_obstacles()
        self.reset_goals()
        return self._get_ob(), self._info

    def render(self) -> None:
        """Rendering the simulation environment.

        As rendering is done rather by the self._render flag,
        only the sleep statement is called here. This speeds up
        the simulation when rendering is not desired.

        """
        #time.sleep(self.dt())

    def close(self) -> None:
        p.disconnect(self._cid)


    def dump(self, file_name: str) -> None:
        with open(file_name, 'wb') as f:
            dill.dump(self,  f)



"""Module for fsd sensor based on lidar."""
import numpy as np
import pybullet
import gymnasium as gym

from urdfenvs.sensors.sensor import Sensor
from urdfenvs.sensors.fsd import FreeSpaceDecomposition
from urdfenvs.urdf_common.helpers import add_shape, quaternion_between_vectors


class FSDSensor(Sensor):
    def __init__(
        self,
        max_radius: float = 1.0,
        number_constraints: int = 10,
        plotting_interval_fsd: int = -1,
        variance: float = 0.0,
        planar_visualization: bool = True,
    ):
        Sensor.__init__(
            self,
            "AbstractFSDSensor",
            variance=variance,
        )
        self._fsd = FreeSpaceDecomposition(
            np.array([0.0, 0.0, 0.0]),
            max_radius=max_radius,
            number_constraints=number_constraints,
        )
        self._plotting_interval_fsd = plotting_interval_fsd
        self._plane_ids = []
        self._planar_visualization = planar_visualization

    def get_observation_space(self, obstacles: dict, goals: dict):
        """Create observation space, all observations should be inside the
        observation space."""
        observation_space = {}
        for i in range(self._fsd._number_constraints):
            observation_space[f"constraint_{i}"] = gym.spaces.Box(
                -np.inf,
                np.inf,
                shape=(4,),
                dtype=float,
            )
        return gym.spaces.Dict({self._name: gym.spaces.Dict(observation_space)})

    def compute_fsd(
        self, point_positions: np.ndarray, center_position: np.ndarray
    ):
        self._fsd.set_position(center_position)
        self._fsd.compute_constraints(point_positions)
        if (
            self._plotting_interval_fsd > 0
            and self._call_counter % self._plotting_interval_fsd == 0
        ):
            if self._planar_visualization:
                self.visualize_constraints()
            else:
                self.visualize_constraints_with_boxes(center_position)
        return self._fsd.asdict()

    def visualize_constraints(self):
        plot_points = self._fsd.get_points()
        pybullet.removeAllUserDebugItems()
        for plot_point in plot_points:
            start_point = (plot_point[0, 0], plot_point[-1, 0], self._height)
            end_point = (plot_point[0, 1], plot_point[-1, 1], self._height)
            pybullet.addUserDebugLine(start_point, end_point)

    def visualize_constraints_with_boxes(self, center_position: np.ndarray):
        constraints = self._fsd.constraints()
        positions = []
        orientations = []
        half_extens = []
        shape_types = []
        for constraint in constraints:
            normal = constraint.normal()
            point = constraint.point()
            vector = np.array([1.0, 0.0, 0.0])
            orientation = quaternion_between_vectors(
                vector, normal, ordering="xyzw",
            )
            orientations.append(orientation)
            shape_types.append(pybullet.GEOM_BOX)
            half_extens.append([0.02, 10.0, 10.0])
            positions.append(point)

        for plane_id in self._plane_ids:
            pybullet.removeBody(plane_id)
        self._plane_ids = []
        visual_shape_id = pybullet.createVisualShapeArray(
            shape_types,
            halfExtents=half_extens,
            visualFramePositions=positions,
            visualFrameOrientations=orientations,
        )
        bullet_id = pybullet.createMultiBody(
            baseMass=0,
            baseCollisionShapeIndex=-1,
            baseVisualShapeIndex=visual_shape_id,
            useMaximalCoordinates=False,
        )
        pybullet.changeVisualShape(
            bullet_id, -1, rgbaColor=[0.0, 0.0, 0.0, 0.1]
        )
        self._plane_ids.append(bullet_id)

        body_id_sphere = add_shape(
            "sphere",
            size=[0.15],
            color=[1.0, 0.0, 0.0, 0.3],
            position=center_position,
            with_collision_shape=False,
        )
        self._plane_ids.append(body_id_sphere)




import pybullet as p
import gymnasium as gym
import numpy as np
import logging

from urdfenvs.urdf_common.generic_robot import GenericRobot


class BicycleModel(GenericRobot):
    """Bicycle model for car like vehicles.

    Attributes
    ----------

    _wheel_radius : float
        The radius of the actuated wheels.
    _spawn_offset : np.ndarray
        The offset by which the initial position must be shifted to align
        observation with that position.
    """

    def __init__(self, n: int, urdf_file: str, mode: str):
        """Constructor for bicyle model robot."""
        super().__init__(n, urdf_file, mode)
        self._wheel_radius: float = None
        self._spawn_offset: np.ndarray = np.array([0.0, 0.0, 0.15])

    def ns(self) -> int:
        """Returns the number of degrees of freedom.

        This is needed as number of actuated joints `_n` is lower that the
        number of degrees of freedom for bycycle models.
        """
        return self.n() + 1

    def reset(
            self,
            pos: np.ndarray,
            vel: np.ndarray,
            mount_position: np.ndarray,
            mount_orientation: np.ndarray,) -> None:
        """ Reset simulation and add robot """
        logging.warning(
            "The argument 'mount_position' and 'mount_orientation' are \
ignored for bicycle models."
        )
        if hasattr(self, "_robot"):
            p.removeBody(self._robot)
        base_orientation = p.getQuaternionFromEuler([0, 0, pos[2]])
        spawn_position = self._spawn_offset
        spawn_position[0:2] += pos[0:2]
        self._robot = p.loadURDF(
            fileName=self._urdf_file,
            basePosition=spawn_position,
            baseOrientation=base_orientation,
            flags=p.URDF_USE_SELF_COLLISION_EXCLUDE_PARENT,
            globalScaling=self._scaling,
        )
        self.set_joint_names()
        self.extract_joint_ids()
        self.read_limits()
        # set base velocity
        self.update_state()
        self._integrated_forward_velocity = vel[0]

    def read_limits(self) -> None:
        self._limit_pos_j = np.zeros((2, self.ns()))
        self._limit_vel_j = np.zeros((2, self.ns()))
        self._limit_tor_j = np.zeros((2, self.n()))
        self._limit_acc_j = np.zeros((2, self.n()))
        self._limit_pos_steering = np.zeros(2)
        joint = self._urdf_robot.robot.joints[self._steering_joints[1] - 1]
        self._limit_pos_steering[0] = joint.limit.lower - 0.1
        self._limit_pos_steering[1] = joint.limit.upper + 0.1
        self._limit_vel_forward_j = np.array([[-40., -10.], [40., 10.]])
        self._limit_pos_j[0, 0:3] = np.array([-1000., -1000., -2 * np.pi])
        self._limit_pos_j[1, 0:3] = np.array([1000., 1000, 2 * np.pi])
        self._limit_vel_j[0, 0:3] = np.array([-40., -40., -10.])
        self._limit_vel_j[1, 0:3] = np.array([40., 40., 10.])
        self.set_acceleration_limits()

    def check_state(self, pos: np.ndarray, vel: np.ndarray) -> tuple:
        """Filters state of the robot and returns a valid state."""

        if not isinstance(pos, np.ndarray) or not pos.size == self.ns():
            pos = np.zeros(self.ns())
        if not isinstance(vel, np.ndarray) or not vel.size == self.n():
            vel = np.zeros(self.n())
        return pos, vel

    def get_observation_space(self) -> gym.spaces.Dict:
        """Gets the observation space for a bicycle model.

        The observation space is represented as a dictonary. `x` and `xdot`
        denote the configuration position and velocity, `vel` for forward and
        angular velocity and `steering` is the current steering position.
        """
        return gym.spaces.Dict(
            {
                "joint_state": gym.spaces.Dict(
                    {
                        "position": gym.spaces.Box(
                            low=self._limit_pos_j[0, :],
                            high=self._limit_pos_j[1, :],
                            dtype=float,
                        ),
                        "steering": gym.spaces.Box(
                            low=self._limit_pos_steering[0],
                            high=self._limit_pos_steering[1],
                            shape=(1,),
                            dtype=float,
                        ),
                        "velocity": gym.spaces.Box(
                            low=self._limit_vel_j[0, :],
                            high=self._limit_vel_j[1, :],
                            dtype=float,
                        ),
                        "forward_velocity": gym.spaces.Box(
                            low=self._limit_vel_forward_j[0, :],
                            high=self._limit_vel_forward_j[1, :],
                            dtype=float,
                        ),
                    }
                )
            }
        )

    def get_velocity_spaces(self) -> tuple:
        """Get observation space and action space when using velocity
        control."""
        ospace = self.get_observation_space()
        uu = self._limit_vel_forward_j[1, :]
        ul = self._limit_vel_forward_j[0, :]
        aspace = gym.spaces.Box(low=ul, high=uu, dtype=float)
        return (ospace, aspace)

    def apply_velocity_action(self, vels: np.ndarray) -> None:
        """Applies velocities to steering and forward motion."""
        p.setJointMotorControl2(
            self._robot,
            self._steering_joints[1],
            controlMode=p.VELOCITY_CONTROL,
            targetVelocity=vels[1],
        )
        pos_wheel_right, _, _, _ = p.getJointState(
            self._robot, self._steering_joints[1]
        )
        p.setJointMotorControl2(
            self._robot,
            self._steering_joints[0],
            controlMode=p.POSITION_CONTROL,
            targetPosition=pos_wheel_right,
        )
        for joint in self._forward_joints:
            p.setJointMotorControl2(
                self._robot,
                joint,
                controlMode=p.VELOCITY_CONTROL,
                targetVelocity=vels[0] / (self._wheel_radius * self._scaling),
            )

    def apply_acceleration_action(self, accs: np.ndarray, dt: float) -> None:
        """Applies acceleration action to the robot.

        The acceleration action relies on integration of the velocity signal
        and applies velocities at the end. The integrated velocities are
        clipped to avoid very large velocities.
        """
        self._integrated_forward_velocity += dt * accs[0]
        actions = np.array([self._integrated_forward_velocity, accs[1]])
        self.apply_velocity_action(actions)

    def apply_torque_action(self, torques: np.ndarray) -> None:
        raise NotImplementedError("Torque action is not available for prius.")

    def correct_base_orientation(self, pos_base: np.ndarray) -> np.ndarray:
        """Corrects base orientation by -pi.

        The orientation observation should be zero when facing positive
        x-direction. Some robot models are rotated by pi. This is corrected
        here. The function also makes sure that the orientation is always
        between -pi and pi.
        """
        pos_base[2] -= np.pi
        if pos_base[2] < -np.pi:
            pos_base[2] += 2 * np.pi
        return pos_base

    def update_state(self) -> None:
        """Updates the robot state.

        The robot state is stored in the dictonary self.state.  There, the key
        x refers to the translational and rotational position in the world
        frame. The key xdot refers to the translation and rotational velocity
        in the world frame. The key vel is the current forward and angular
        velocity. For a bicycle model we additionally store information about
        the sterring behind the key steering.
        """
        # base position
        link_state = p.getLinkState(self._robot, 0, computeLinkVelocity=1)
        pos_base = np.array(
            [
                link_state[0][0],
                link_state[0][1],
                p.getEulerFromQuaternion(link_state[1])[2],
            ]
        )
        self.correct_base_orientation(pos_base)
        vel_base = np.array(
            [link_state[6][0], link_state[6][1], link_state[7][2]]
        )
        # wheel velocities
        vel_wheels = p.getJointStates(self._robot, self._forward_joints[2:4])
        v_right = vel_wheels[0][1]
        v_left = vel_wheels[1][1]
        vel = np.array(
            [
                0.5 * (v_right + v_left) * self._scaling * self._wheel_radius,
                vel_base[2],
            ]
        )
        pos, _, _, _ = p.getJointState(self._robot, self._steering_joints[1])
        steering_pos = np.array([pos])
        self.state = {
            "joint_state": {
                "position": pos_base,
                "forward_velocity": vel,
                "velocity": vel_base,
                "steering": steering_pos,
            }
        }



"""Abstract class for sensor."""
from typing import List
from abc import abstractmethod


class Sensor(object):
    """Abstract sensor class.

    This class serves as a blueprint for sensors. Every sensor must
    inherit from this class and implement the abstract methods.
    """

    def __init__(
        self, name, variance: float = 0.0, plotting_interval: int = -1
    ):
        self._name = name
        self._variance = variance
        self._plotting_interval = plotting_interval
        self._call_counter = plotting_interval - 1

    def name(self):
        return self._name

    @abstractmethod
    def get_observation_size(self):
        pass

    @abstractmethod
    def get_observation_space(self, obstacles: dict, goals: dict):
        pass

    @abstractmethod
    def sense(self, robot, obstacles: dict, goals: dict, t: float):
        pass



import logging
from typing import Tuple, Optional, List

import numpy as np
import pybullet


def quaternion_between_vectors(v1, v2, ordering="wxyz"):
    quaternion = np.zeros(4)
    a_part = np.cross(v1, v2)
    if ordering == "wxyz":
        quaternion[0] = np.sqrt(
            (np.linalg.norm(v1) ** 2) * (np.linalg.norm(v2) ** 2)
        ) + np.dot(v1, v2)
        quaternion[1:4] = a_part
    elif ordering == "xyzw":
        quaternion[0:3] = a_part
        quaternion[3] = np.sqrt(
            (np.linalg.norm(v1) ** 2) * (np.linalg.norm(v2) ** 2)
        ) + np.dot(v1, v2)
    else:
        raise InvalidQuaternionOrderError(
            f"Order {ordering} is not permitted, options are 'xyzw', and 'wxyz'"
        )
    normed_quat = quaternion / np.linalg.norm(quaternion)
    return normed_quat


class LinkIdNotFoundError(Exception):
    pass


def extract_link_id(robot, link_name: str):
    number_links = pybullet.getNumJoints(robot)
    joint_names = []
    for i in range(number_links):
        joint_name = pybullet.getJointInfo(robot, i)[1].decode("UTF-8")
        joint_names.append(joint_name)
        if joint_name == link_name:
            return i
    raise LinkIdNotFoundError(
        f"Link with name {link_name} not found. "
        f"Possible links are {joint_names}"
    )


class InvalidQuaternionOrderError(Exception):
    pass


def quaternion_to_rotation_matrix(
    quaternion: np.ndarray, ordering: str = "wxyz"
) -> np.ndarray:
    # Normalize the quaternion if needed
    quaternion /= np.linalg.norm(quaternion)

    if ordering == "wxyz":
        w, x, y, z = quaternion
    elif ordering == "xyzw":
        x, y, z, w = quaternion
    else:
        raise InvalidQuaternionOrderError(
            f"Order {ordering} is not permitted, options are 'xyzw', and 'wxyz'"
        )
    rotation_matrix = np.array(
        [
            [
                1 - 2 * y**2 - 2 * z**2,
                2 * x * y - 2 * w * z,
                2 * x * z + 2 * w * y,
            ],
            [
                2 * x * y + 2 * w * z,
                1 - 2 * x**2 - 2 * z**2,
                2 * y * z - 2 * w * x,
            ],
            [
                2 * x * z - 2 * w * y,
                2 * y * z + 2 * w * x,
                1 - 2 * x**2 - 2 * y**2,
            ],
        ]
    )

    return rotation_matrix


def get_transformation_matrix(
    quaternion: np.ndarray, translation: np.ndarray
) -> np.ndarray:
    rotation = quaternion_to_rotation_matrix(quaternion, ordering="xyzw")

    transformation_matrix = np.eye(4)
    transformation_matrix[:3, :3] = rotation
    transformation_matrix[:3, 3] = translation

    return transformation_matrix


def matrix_to_quaternion(matrix, ordering="wxyz") -> Tuple[float]:
    """
    Convert a 4x4 transformation matrix to a quaternion.

    Parameters:
        matrix (numpy.ndarray): The 4x4 transformation matrix.

    Returns:
        numpy.ndarray: The quaternion representation (w, x, y, z).
    """

    # Extract the rotation matrix from the transformation matrix
    rotation_matrix = matrix[:3, :3]
    translation = matrix[:3, 3]

    # Calculate the trace of the rotation matrix
    trace = np.trace(rotation_matrix)

    if trace > 0:
        # The quaternion calculation when the trace is positive
        s = np.sqrt(trace + 1.0) * 2
        w = 0.25 * s
        x = (rotation_matrix[2, 1] - rotation_matrix[1, 2]) / s
        y = (rotation_matrix[0, 2] - rotation_matrix[2, 0]) / s
        z = (rotation_matrix[1, 0] - rotation_matrix[0, 1]) / s
    elif (rotation_matrix[0, 0] > rotation_matrix[1, 1]) and (
        rotation_matrix[0, 0] > rotation_matrix[2, 2]
    ):
        # The quaternion calculation when the trace is largest along x-axis
        s = (
            np.sqrt(
                1.0
                + rotation_matrix[0, 0]
                - rotation_matrix[1, 1]
                - rotation_matrix[2, 2]
            )
            * 2
        )
        w = (rotation_matrix[2, 1] - rotation_matrix[1, 2]) / s
        x = 0.25 * s
        y = (rotation_matrix[0, 1] + rotation_matrix[1, 0]) / s
        z = (rotation_matrix[0, 2] + rotation_matrix[2, 0]) / s
    elif rotation_matrix[1, 1] > rotation_matrix[2, 2]:
        # The quaternion calculation when the trace is largest along y-axis
        s = (
            np.sqrt(
                1.0
                + rotation_matrix[1, 1]
                - rotation_matrix[0, 0]
                - rotation_matrix[2, 2]
            )
            * 2
        )
        w = (rotation_matrix[0, 2] - rotation_matrix[2, 0]) / s
        x = (rotation_matrix[0, 1] + rotation_matrix[1, 0]) / s
        y = 0.25 * s
        z = (rotation_matrix[1, 2] + rotation_matrix[2, 1]) / s
    else:
        # The quaternion calculation when the trace is largest along z-axis
        s = (
            np.sqrt(
                1.0
                + rotation_matrix[2, 2]
                - rotation_matrix[0, 0]
                - rotation_matrix[1, 1]
            )
            * 2
        )
        w = (rotation_matrix[1, 0] - rotation_matrix[0, 1]) / s
        x = (rotation_matrix[0, 2] + rotation_matrix[2, 0]) / s
        y = (rotation_matrix[1, 2] + rotation_matrix[2, 1]) / s
        z = 0.25 * s

    quaternion = np.array([1, 0, 0, 0])
    if ordering == "wxyz":
        quaternion = np.array([w, x, y, z])
    elif ordering == "xyzw":
        quaternion = np.array([x, y, z, w])
    else:
        raise InvalidQuaternionOrderError(
            f"Order {ordering} is not permitted, options are 'xyzw', and 'wxyz'"
        )

    return translation, quaternion


def add_shape(
    shape_type: str,
    size: list,
    color: Optional[List[float]] = None,
    movable: bool = False,
    orientation: Optional[Tuple[float]] = None,
    position: Optional[Tuple[float]] = None,
    scaling: float = 1.0,
    urdf: Optional[str] = None,
    with_collision_shape: bool = True,
) -> int:

    mass = float(movable)
    if color is None:
        color = [1.0, 1.0, 1.0, 1.0]
    if orientation is None:
        base_orientation = (0.0, 0.0, 0.0, 1.0)
    else:
        base_orientation = orientation
    if position is None:
        base_position = (0.0, 0.0, 0.0)
    else:
        base_position = position
    if shape_type in ["sphere", "splineSphere", "analyticSphere"]:
        shape_id = pybullet.createCollisionShape(
            pybullet.GEOM_SPHERE, radius=size[0]
        )
        visual_shape_id = pybullet.createVisualShape(
            pybullet.GEOM_SPHERE,
            rgbaColor=color,
            specularColor=[1.0, 0.5, 0.5],
            radius=size[0],
        )

    elif shape_type == "box":
        half_extens = [s / 2 for s in size]
        base_position = tuple(base_position[i] - size[i] for i in range(3))
        shape_id = pybullet.createCollisionShape(
            pybullet.GEOM_BOX, halfExtents=half_extens
        )
        visual_shape_id = pybullet.createVisualShape(
            pybullet.GEOM_BOX,
            rgbaColor=color,
            specularColor=[1.0, 0.5, 0.5],
            halfExtents=half_extens,
        )

    elif shape_type == "cylinder":
        shape_id = pybullet.createCollisionShape(
            pybullet.GEOM_CYLINDER, radius=size[0], height=size[1]
        )
        visual_shape_id = pybullet.createVisualShape(
            pybullet.GEOM_CYLINDER,
            rgbaColor=color,
            specularColor=[1.0, 0.5, 0.5],
            radius=size[0],
            length=size[1],
        )

    elif shape_type == "capsule":
        shape_id = pybullet.createCollisionShape(
            pybullet.GEOM_CAPSULE, radius=size[0], height=size[1]
        )
        visual_shape_id = pybullet.createVisualShape(
            pybullet.GEOM_CAPSULE,
            rgbaColor=color,
            specularColor=[1.0, 0.5, 0.5],
            radius=size[0],
            length=size[1],
        )
    elif shape_type == "urdf":
        shape_id = pybullet.loadURDF(
            fileName=urdf, basePosition=base_position, globalScaling=scaling
        )
        return shape_id
    else:
        logging.warning("Unknown shape type: {shape_type}, aborting...")
        return -1
    if not with_collision_shape:
        shape_id = -1
    bullet_id = pybullet.createMultiBody(
        baseMass=mass,
        baseCollisionShapeIndex=shape_id,
        baseVisualShapeIndex=visual_shape_id,
        basePosition=base_position,
        baseOrientation=base_orientation,
    )
    return bullet_id



from pynput import keyboard
from pynput.keyboard import Key
import numpy as np
import warnings

# reserved keys are detected, but not executed by the gym environment
reserved_keys = [
    Key.esc,
    "q",
    "w",
    "e",
    "r",
    "t",
    "y",
    "u",
    "i",
    "o",
    "p",
    "a",
    "s",
    "d",
    "f",
    "g",
    "h",
    "j",
    "k",
    "l",
    "z",
    "x",
    "c",
    "v",
    "b",
    "n",
    "m",
    "<",
    ",",
    ">",
    ".",
    "/",
    "?",
    ":",
    ";",
    "'",
    "[",
    "]",
    "{",
    "}",
    '"',
    "|",
    "`",
    "~",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    "0",
    "_",
    "-",
    "=",
    "+",
    "!",
    "@",
    "#",
    "$",
    "%",
    "^",
    "&",
    "*",
    "(",
    ")",
]


class Responder:
    """
    The Responder class listens to keyboard input from the user and updates the
    class variable action.  An action can be pulled with the respond function.
    Using this pattern has 2 advantages:
        * keyboard listener frequency does not depend on game loop frequency
        * actions are only send if they are going to be processed

    Usage
    initializing and set up:
        * Pipe between parent and child processes
        * Main loop(parent_connection) containing the main loop
        * Responder(child_connection) object

    In the main loop ask for an action and wait for the responder to
    respond as follows:
        parent_connection.send({"request_action": True})
        keyboard_data = parent_connection.recv()
    """

    def __init__(self, child_conn):
        """
        :param child_conn: Pipe connection to the parent process
        """
        self._conn = child_conn
        self._default_action = None
        self._action = None

    def _on_press_outer(self, custom_on_press=None):
        """
        Updating the class variable action on key press
        :param custom_on_press: Dictionary with key action bindings
        """
        if custom_on_press is None:
            # default action keybindings
            def on_press(key):
                if key == Key.up:
                    self.action = np.array([1.0, 0.0])
                if key == Key.down:
                    self.action = np.array([-1.0, 0.0])
                if key == Key.left:
                    self.action = np.array([0.0, 1.0])
                if key == Key.right:
                    self.action = np.array([0.0, -1.0])

        else:
            # custom action keybindings
            def on_press(key):
                # print("key pressed: {}".format(key))
                for custom_key in custom_on_press:
                    if key == custom_key:
                        self.action = custom_on_press[custom_key]

        return on_press

    def _on_release_outer(self, custom_on_release=None):
        """
        Updating the class variable action to default action on key release
        :param custom_on_release: Dictionary with key action bindings
        """
        if custom_on_release is None:
            # default action keybindings
            def on_press(key):
                self.action = self.default_action

        else:
            # custom action keybindings
            def on_press(key):
                for custom_key in custom_on_release:
                    if key == custom_key:
                        self.action = custom_on_release[custom_key]

        return on_press

    def respond(self):
        """Respond to request with the latest action"""

        # receive request
        request = self.conn.recv()
        if request["request_action"]:
            # send action
            self.conn.send({"action": self._action})
        elif request["kill_child"]:
            raise Exception
        else:
            raise Exception("Cannot handle action request")

    def setup(
        self,
        default_action=np.array([0.0, 0.0]),
        custom_on_press=None,
        custom_on_release=None,
    ):
        """
        Setup responder, optionally set defaultaction and custom action
        keybindings

        Parameters
        ----------

        default_action: np.ndarray
            Default action when no action is specified
        custom_on_press:
            Custom on press callback function for key bindings
        custom_on_release:
            Custom on release callback function for key bindings
        """
        self.default_action = default_action
        self.action = default_action

        # warn if a reserved key is in the custom_on_press keys
        if custom_on_press is not None:
            if (
                len(
                    [
                        key
                        for key in reserved_keys
                        if key in custom_on_press.keys()
                    ]
                )
                > 0
            ):
                warnings.warn("reserved key used for control")

        # warn if a reserved key is in the custom_on_release keys
        if custom_on_release is not None:
            if (
                len(
                    [
                        key
                        for key in reserved_keys
                        if key in custom_on_release.keys()
                    ]
                )
                > 0
            ):
                warnings.warn("reserved key used for control")

        # setup keyboard listener
        listener = keyboard.Listener(
            on_press=self._on_press_outer(custom_on_press),
            on_release=self._on_release_outer(custom_on_release),
        )
        # start listening to keyboard input
        listener.start()

    def start(self, p):
        """
        start with responding to keyboard input

        :param p: Parent process
        """

        # while parent process is alive, keep on responding
        while p.is_alive():
            try:
                self.respond()
            except Exception:
                return

    # getters and setters
    @property
    def conn(self):
        return self._conn

    @property
    def default_action(self):
        return self._default_action

    @default_action.setter
    def default_action(self, value):
        self._default_action = value

    @property
    def action(self):
        return self._action

    @action.setter
    def action(self, value):
        self._action = value



from urdfenvs.sensors.sensor import Sensor
import numpy as np
import pybullet as p
from gymnasium import spaces


class FullSensor(Sensor):
    def __init__(
        self, goal_mask: list, obstacle_mask: list, variance: float = 0.0
    ):
        super().__init__("FullSensor", variance=variance)
        self._obstacle_mask = obstacle_mask
        self._goal_mask = goal_mask

    def _reset(self):
        pass

    def observation_size(self) -> tuple:
        return 1, 2

    def get_observation_space(self, obstacles: dict, goals: dict):
        observation_space = {}
        observation_space_obstacles = {}
        for obst_id, obstacle in obstacles.items():
            observation_space_obstacle = {}
            if "position" in self._obstacle_mask:
                observation_space_obstacle["position"] = spaces.Box(
                    low=np.array([-50, -50, -50]),
                    high=np.array([50, 50, 50]),
                    dtype=float,
                )
            if "velocity" in self._obstacle_mask:
                observation_space_obstacle["velocity"] = spaces.Box(
                    low=np.array([-50, -50, -50]),
                    high=np.array([50, 50, 50]),
                    dtype=float,
                )
            if "acceleration" in self._obstacle_mask:
                observation_space_obstacle["acceleration"] = spaces.Box(
                    low=np.array([-5, -5, -5]),
                    high=np.array([5, 5, 5]),
                    dtype=float,
                )
            if "type" in self._obstacle_mask:
                observation_space_obstacle["type"] = spaces.Discrete(128)
            if "size" in self._obstacle_mask:
                low_limit_size = [
                    0,
                ] * len(obstacle.size())
                high_limit_size = [
                    50,
                ] * len(obstacle.size())
                observation_space_obstacle["size"] = spaces.Box(
                    low=np.array(low_limit_size),
                    high=np.array(high_limit_size),
                    dtype=float,
                )
            observation_space_obstacles[obst_id] = spaces.Dict(
                observation_space_obstacle
            )
        if observation_space_obstacles:
            observation_space["obstacles"] = spaces.Dict(
                observation_space_obstacles
            )
        observation_space_goals = {}
        for goal_id, goal in goals.items():
            observation_space_goal = {}
            if "position" in self._goal_mask:
                observation_space_goal["position"] = spaces.Box(
                    low=np.array([-50, -50, -50]),
                    high=np.array([50, 50, 50]),
                    dtype=float,
                )
            if "velocity" in self._goal_mask:
                observation_space_goal["velocity"] = spaces.Box(
                    low=np.array([-50, -50, -50]),
                    high=np.array([50, 50, 50]),
                    dtype=float,
                )
            if "acceleration" in self._goal_mask:
                observation_space_goal["acceleration"] = spaces.Box(
                    low=np.array([-5, -5, -5]),
                    high=np.array([5, 5, 5]),
                    dtype=float,
                )
            if "weight" in self._goal_mask:
                observation_space_goal["weight"] = spaces.Box(
                    low=np.array([0]),
                    high=np.array([10]),
                    dtype=float,
                )
            if "is_primary_goal" in self._goal_mask:
                observation_space_goal["is_primary_goal"] = spaces.Box(
                    low=np.array([False], dtype=bool),
                    high=np.array([True], dtype=bool),
                    dtype=bool,
                )
            observation_space_goals[goal_id] = spaces.Dict(
                observation_space_goal
            )
        if observation_space_goals:
            observation_space["goals"] = spaces.Dict(observation_space_goals)
        return spaces.Dict({self._name: spaces.Dict(observation_space)})

    def sense(self, robot, obstacles: dict, goals: dict, t: float):
        sensor_observation = {}

        observations = {}
        for obst_id, obstacle in obstacles.items():
            observation = {}
            for mask_item in self._obstacle_mask:
                if mask_item == "position":
                    value, _ = p.getBasePositionAndOrientation(obst_id)

                elif mask_item == "velocity":
                    value, _ = p.getBaseVelocity(obst_id)

                else:
                    try:
                        value = getattr(obstacle, mask_item)(t=t)
                    except TypeError:
                        value = getattr(obstacle, mask_item)()
                if isinstance(value, float):
                    value = [value]
                if isinstance(value, str):
                    observation[mask_item] = np.array([ord(c) for c in value])
                else:
                    observation[mask_item] = np.random.normal(
                        np.array(value), self._variance
                    ).astype("float32")
            observations[obst_id] = observation

        if observations:
            sensor_observation["obstacles"] = observations

        observations = {}
        for obst_id, goal in goals.items():
            observation = {}
            for mask_item in self._goal_mask:
                if mask_item == "position":
                    value, _ = p.getBasePositionAndOrientation(obst_id)

                elif mask_item == "velocity":
                    value, _ = p.getBaseVelocity(obst_id)

                else:
                    try:
                        value = getattr(goal, mask_item)(t=t)
                    except TypeError:
                        value = getattr(goal, mask_item)()
                if isinstance(value, float):
                    value = [value]
                if isinstance(value, bool):
                    observation[mask_item] = np.array([value])
                else:
                    observation[mask_item] = np.random.normal(
                        np.array(value), self._variance
                    ).astype("float32")
            observations[obst_id] = observation

        if observations:
            sensor_observation["goals"] = observations

        return sensor_observation



import os
import numpy as np

from urdfenvs.urdf_common.bicycle_model import BicycleModel


class Prius(BicycleModel):
    """Prius robot model.

    Attributes
    ---------
    _scaling: float
        The size scaling in which the urdf should be spawned.
        This also effects the dynamics of the system.

    """
    def __init__(self, mode: str):
        n = 2
        urdf_file = os.path.join(os.path.dirname(__file__), 'prius.urdf')
        self._scaling: float = 0.3
        super().__init__(n, urdf_file, mode)
        self._wheel_radius = 0.31265
        self._wheel_distance = 0.494
        self._spawn_offset: np.ndarray = np.array([-0.435, 0.0, 0.05])

    def set_joint_names(self):
        """Set joint indices.

        For the bicycle model robots, the steering joints and the forward joints
        are determined.
        """
        self._robot_joints = [2, 4]
        self._steering_joints = [2, 4]
        self._forward_joints = [3, 5, 6, 7]

    def set_acceleration_limits(self):
        acc_limit = np.array([1.0, 1.0])
        self._limit_acc_j[0, :] = -acc_limit
        self._limit_acc_j[1, :] = acc_limit

    def check_state(self, pos, vel):
        if (
            not isinstance(pos, np.ndarray)
            or not pos.size == self.n() + 1
        ):
            pos = np.zeros(self.n() + 1)
        if not isinstance(vel, np.ndarray) or not vel.size == self.n():
            vel = np.zeros(self.n())
        return pos, vel



import os
from enum import Enum
from abc import ABC, abstractmethod
from typing import List
import pybullet as p
import gymnasium as gym
import numpy as np
import yourdfpy
import urdfenvs
from urdfenvs.sensors.sensor import Sensor

class ControlMode(Enum):
    torque = 'tor'
    acceleration = 'acc'
    velocity = 'vel'

class GenericRobot(ABC):
    """GenericRobot."""
    _castor_wheels = []

    def __init__(self, n: int, urdf_file: str, mode=ControlMode.velocity):
        """Constructor for generic robot.

        Parameters
        ----------

        n: int : Degrees of freedom of the robot
        urdf_file: str : Full path to urdf file
        """
        # search for urdf in package if not found in cwd
        if not os.path.exists(urdf_file):
            asset_dir = urdfenvs.__path__[0] + "/assets"
            asset_urdf = None
            for root, _, files in os.walk(asset_dir):
                for file in files:
                    if file == urdf_file:
                        asset_urdf = os.path.join(root, file)
            if asset_urdf is None:
                raise Exception(f"the request urdf {urdf_file} can not be found")
            self._urdf_file = asset_urdf
        else:
            self._urdf_file = urdf_file
        self._sensors: List[Sensor] = []
        self._urdf_robot = yourdfpy.urdf.URDF.load(self._urdf_file)
        self._mode = ControlMode(mode)
        self.set_degrees_of_freedom(n)
        self._link_names = []
        self.set_joint_names()
        self.extract_joint_ids()
        self.read_limits()

    def set_degrees_of_freedom(self, n):
        if n > 0:
            self._n = n
        else:
            self._n: int = self._urdf_robot.num_actuated_joints

    def n(self) -> int:
        return self._n

    def ns(self) -> int:
        return self.n()

    @abstractmethod
    def reset(
            self,
            pos: np.ndarray,
            vel: np.ndarray,
            mount_position: np.ndarray,
            mount_orientation: np.ndarray,) -> None:
        """Resets the robot to an initial state.

        Parameters
        ----------
        pos: np.ndarray:
            Initial joint positions
        vel: np.ndarray:
            Initial joint velocities
        mount_position: np.ndarray:
            Mounting position
        mount_orientation: np.ndarray:
            Mounting orientation

        """
        pass

    @abstractmethod
    def set_joint_names(self) -> None:
        """Sets joint indices for urdf parsing.

        Input the names of joints manually.

        """
        pass

    @abstractmethod
    def read_limits(self) -> None:
        """Read and set the joint limits."""
        pass

    @abstractmethod
    def set_acceleration_limits(self):
        pass

    def extract_joint_ids(self) -> None:
        """Automated extraction of joint ids

        Extract joint ids by the joint names.

        """
        if not hasattr(self, "_joint_names"):
            return
        self._urdf_joints = []
        for i, joint_name in enumerate(self._urdf_robot.joint_names):
            if joint_name in self._joint_names:
                self._urdf_joints.append(i)
        if hasattr(self, "_robot"):
            self._robot_joints = []
            self._castor_joints = []
            num_joints = p.getNumJoints(self._robot)
            for name in self._joint_names:
                for i in range(num_joints):
                    joint_info = p.getJointInfo(self._robot, i)
                    joint_name = joint_info[1].decode("UTF-8")
                    link_name = joint_info[12].decode("UTF-8")
                    if joint_name == name:
                        self._robot_joints.append(i)
                    self._link_names.append(link_name)
                for i in range(num_joints):
                    joint_info = p.getJointInfo(self._robot, i)
                    joint_name = joint_info[1].decode("UTF-8")
                    if joint_name in self._castor_wheels:
                        self._castor_joints.append(i)


    def get_observation_space(self) -> gym.spaces.Dict:
        """Get observation space."""
        return gym.spaces.Dict(
            {
                "joint_state": gym.spaces.Dict(
                    {
                        "position": gym.spaces.Box(
                        low=self._limit_pos_j[0, :],
                        high=self._limit_pos_j[1, :],
                        dtype=float,
                    ),
                        "velocity": gym.spaces.Box(
                            low=self._limit_vel_j[0, :],
                            high=self._limit_vel_j[1, :],
                            dtype=float,
                        ),
                    }
                )
            }
        )

    def check_state(self, pos: np.ndarray, vel: np.ndarray) -> tuple:
        """Filters state of the robot and returns a valid state."""

        if not isinstance(pos, np.ndarray) or not pos.size == self.n():
            pos = np.zeros(self.n())
        if not isinstance(vel, np.ndarray) or not vel.size == self.n():
            vel = np.zeros(self.n())
        return pos, vel


    def get_torque_spaces(self) -> tuple:
        """Get observation space and action space when using torque control."""
        ospace = self.get_observation_space()
        uu = self._limit_tor_j[1, :]
        ul = self._limit_tor_j[0, :]
        aspace = gym.spaces.Box(low=ul, high=uu, dtype=float)
        return (ospace, aspace)

    def get_velocity_spaces(self) -> tuple:
        """Get observation space and action space when using velocity
        control."""
        ospace = self.get_observation_space()
        uu = self._limit_vel_j[1, :]
        ul = self._limit_vel_j[0, :]
        aspace = gym.spaces.Box(low=ul, high=uu, dtype=float)
        return (ospace, aspace)

    def get_acceleration_spaces(self) -> tuple:
        """Get observation space and action space when using acceleration
        control."""
        ospace = self.get_observation_space()
        uu = self._limit_acc_j[1, :]
        ul = self._limit_acc_j[0, :]
        aspace = gym.spaces.Box(low=ul, high=uu, dtype=float)
        return (ospace, aspace)

    def disable_velocity_control(self):
        """Disables velocity control for all controlled joints.

        By default, pybullet uses velocity control. This has to be disabled if
        torques should be directly controlled.  See
        func:`~urdfenvs.urdfCommon.generic_robot.generic_rob
        ot.apply_torque_action`
        """
        self._friction = 0.0
        for i in range(self._n):
            p.setJointMotorControl2(
                self._robot,
                jointIndex=self._robot_joints[i],
                controlMode=p.VELOCITY_CONTROL,
                force=self._friction,
            )

    @abstractmethod
    def apply_torque_action(self, torques) -> None:
        pass

    @abstractmethod
    def apply_velocity_action(self, vels) -> None:
        pass

    @abstractmethod
    def apply_acceleration_action(self, accs) -> None:
        pass

    def apply_action(self, action, dt=None) -> None:
        if self._mode == ControlMode.torque:
            self.apply_torque_action(action)
        elif self._mode == ControlMode.velocity:
            self.apply_velocity_action(action)
        elif self._mode == ControlMode.acceleration:
            self.apply_acceleration_action(action, dt)
        else:
            raise Exception(f"ControlMode {self._mode} not implemented")

    def get_spaces(self):
        if self._mode == ControlMode.torque:
            return self.get_torque_spaces()
        elif self._mode == ControlMode.velocity:
            return self.get_velocity_spaces()
        elif self._mode == ControlMode.acceleration:
            return self.get_acceleration_spaces()
        else:
            raise Exception(f"ControlMode {self._mode} not implemented")

    @abstractmethod
    def update_state(self) -> None:
        """Updates the state of the robot.

        This function reads current joint position and velocities from the
        physics engine.

        """
        pass

    def sense(self, obstacles: dict, goals: dict, t: float) -> None:
        """Updates the sensing of the robot's sensors."""
        self.sensor_observation = {}
        for sensor in self._sensors:
            self.sensor_observation[sensor.name()] = sensor.sense(self._robot, obstacles, goals, t)
            #self.sensor_observation.update(sensor.sense(self._robot, obst_ids, goal_ids))

    def get_observation(self, obstacles: dict, goals: dict, t: float) -> dict:
        """Updates all observation and concatenate joint states and sensor
        observation."""
        self.update_state()
        self.sense(obstacles, goals, t)
        return {**self.state, **self.sensor_observation}

    def add_sensor(self, sensor: Sensor) -> int:
        """Adds sensor to the robot."""
        self._sensors.append(sensor)
        return sensor.get_observation_size()

    def sensors(self) -> list:
        return self._sensors



import logging
from copy import deepcopy
from typing import List
import pybullet as p
import gymnasium as gym
import numpy as np

from urdfenvs.urdf_common.generic_robot import ControlMode, GenericRobot

class FacingDirectionUndefinedError(Exception):
    pass


class DifferentialDriveRobot(GenericRobot):
    """Differential drive robot.

    Attributes
    ----------

    _wheel_radius : float
        The radius of the actuated wheels.
    _wheel_distance : float
        The distance between the actuated wheels.
    _spawn_offset : np.ndarray
        The offset by which the initial position must be shifted to align
        observation with that position.
    """
    def __init__(
            self,
            n: int,
            urdf_file: str,
            mode: ControlMode,
            actuated_wheels: List[str],
            castor_wheels: List[str],
            wheel_radius: float,
            wheel_distance: float,
            spawn_offset: np.ndarray = np.array([0.0, 0.0, 0.15]),
            spawn_rotation: float = 0.0,
            facing_direction: str = 'x',
            not_actuated_joints: List[str] = []
    ):
        """Constructor for differential drive robots."""
        self._number_actuated_axes = int(len(actuated_wheels)/2)
        self._spawn_rotation = deepcopy(spawn_rotation)
        self._spawn_offset = deepcopy(spawn_offset)
        self._castor_wheels = castor_wheels
        self._actuated_wheels = actuated_wheels
        self._wheel_radius = wheel_radius
        self._wheel_distance = wheel_distance
        self._not_actuated_joints = not_actuated_joints
        self._facing_direction = facing_direction
        super().__init__(n, urdf_file, mode)

    def set_degrees_of_freedom(self, n):
        if n > 0:
            self._n = n
        else:
            self._n: int = self._urdf_robot.num_actuated_joints - len(self._actuated_wheels) - len(self._castor_wheels) - len(self._not_actuated_joints) + 2
    def ns(self) -> int:
        """Returns the number of degrees of freedom.

        This is needed as number of actuated joints `_n` is lower that the
        number of degrees of freedom for differential drive robots.
        """
        return self.n() + 1

    def get_velocity_spaces(self) -> tuple:
        """Get observation space and action space when using velocity
        control.

        Overrides velocity spaces from default, because a differential drive has limits in x,y and
        theta direction, while the action space should be limited to the forward and angular velocity."""
        ospace = self.get_observation_space()
        uu = np.concatenate(
            (self._limit_vel_forward_j[1, :], self._limit_vel_j[1, 3:]), axis=0
        )
        ul = np.concatenate(
            (self._limit_vel_forward_j[0, :], self._limit_vel_j[0, 3:]), axis=0
        )
        aspace = gym.spaces.Box(low=ul, high=uu, dtype=float)
        return (ospace, aspace)

    def reset(
            self,
            pos: np.ndarray,
            vel: np.ndarray,
            mount_position: np.ndarray,
            mount_orientation: np.ndarray,) -> None:
        """ Reset simulation and add robot """
        logging.warning(
            "The argument 'mount_position' and 'mount_orientation' are \
ignored for differential drive robots."
        )
        if hasattr(self, "_robot"):
            p.removeBody(self._robot)
        base_orientation = p.getQuaternionFromEuler([0, 0, self._spawn_rotation + pos[2]])
        spawn_position = self._spawn_offset
        spawn_position[0:2] += pos[0:2]
        self._robot = p.loadURDF(
            fileName=self._urdf_file,
            basePosition=spawn_position,
            baseOrientation=base_orientation,
            flags=p.URDF_USE_SELF_COLLISION_EXCLUDE_PARENT,
        )
        self.set_joint_names()
        self.extract_joint_ids()
        self.read_limits()
        # Joint indices as found by p.getJointInfo()
        # set castor wheel friction to zero
        for i in self._castor_joints:
            p.setJointMotorControl2(
                self._robot,
                jointIndex=i,
                controlMode=p.VELOCITY_CONTROL,
                force=0.0,
            )
        for i in self._castor_joints:
            p.changeDynamics(self._robot, i, lateralFriction=0)
        for i in range(2, self._n):
            p.resetJointState(
                self._robot,
                self._robot_joints[i],
                pos[i + 1],
                targetVelocity=vel[i],
            )
        # set base velocity
        self.update_state()
        self._integrated_velocities = vel

    def check_state(self, pos: np.ndarray, vel: np.ndarray) -> tuple:
        """Filters state of the robot and returns a valid state."""

        if not isinstance(pos, np.ndarray) or not pos.size == self.ns():
            pos = np.zeros(self.ns())
        if not isinstance(vel, np.ndarray) or not vel.size == self.n():
            vel = np.zeros(self.n())
        return pos, vel

    def read_limits(self) -> None:
        """ Set position, velocity, acceleration
        and motor torque lower en upper limits """
        self._limit_pos_j = np.zeros((2, self.ns()))
        self._limit_vel_j = np.zeros((2, self.ns()))
        self._limit_tor_j = np.zeros((2, self.n()))
        self._limit_acc_j = np.zeros((2, self.n()))
        for i in range(self.n()):
            joint = self._urdf_robot.robot.joints[self._urdf_joints[i]]
            self._limit_tor_j[0, i] = -joint.limit.effort
            self._limit_tor_j[1, i] = joint.limit.effort
            if i >= 2:
                self._limit_pos_j[0, i + 1] = joint.limit.lower
                self._limit_pos_j[1, i + 1] = joint.limit.upper
                self._limit_vel_j[0, i + 1] = -joint.limit.velocity
                self._limit_vel_j[1, i + 1] = joint.limit.velocity
        self._limit_vel_forward_j = np.array([[-4.0, -10.0], [4.0, 10.0]])
        self._limit_pos_j[0, 0:3] = np.array([-10.0, -10.0, -2 * np.pi])
        self._limit_pos_j[1, 0:3] = np.array([10.0, 10.0, 2 * np.pi])
        self._limit_vel_j[0, 0:3] = np.array([-4.0, -4.0, -10.0])
        self._limit_vel_j[1, 0:3] = np.array([4.0, 4.0, 10.0])
        self.set_acceleration_limits()

    def get_observation_space(self) -> gym.spaces.Dict:
        """Gets the observation space for a differential drive robot.

        The observation space is represented as a dictionary.
        `join_state` containing:
        `position` the concatenated positions of joints
        in their local configuration space.
        `velocity` the concatenated velocities of joints
        in their local configuration space.
        `forward_velocity` the forward velocity of the robot.
        """
        return gym.spaces.Dict(
            {
                "joint_state": gym.spaces.Dict(
                    {
                        "position": gym.spaces.Box(
                            low=self._limit_pos_j[0, :],
                            high=self._limit_pos_j[1, :],
                            dtype=float,
                        ),
                        "velocity": gym.spaces.Box(
                            low=self._limit_vel_j[0, :],
                            high=self._limit_vel_j[1, :],
                            dtype=float,
                        ),
                        "forward_velocity": gym.spaces.Box(
                            low=np.array([self._limit_vel_forward_j[0][0]]),
                            high=np.array([self._limit_vel_forward_j[1][0]]),
                            shape=(1,),
                            dtype=float,
                        ),
                    }
                )
            }
        )

    def apply_torque_action(self, torques: np.ndarray) -> None:
        """Applies torque action to the arm joints of the robot.

        Torque control is not available for the base at the moment.
        """
        for i in range(2, self._n):
            p.setJointMotorControl2(
                self._robot,
                self._robot_joints[i],
                controlMode=p.TORQUE_CONTROL,
                force=torques[i],
            )

    def apply_acceleration_action(self, accs: np.ndarray, dt: float) -> None:
        """Applies acceleration action to the robot.

        The acceleration action relies on integration of the velocity signal
        and applies velocities at the end. The integrated velocities are
        clipped to avoid very large velocities.
        """
        self._integrated_velocities += dt * accs
        self._integrated_velocities[0] = np.clip(
            self._integrated_velocities[0],
            0.7 * self._limit_vel_forward_j[0, 0],
            0.7 * self._limit_vel_forward_j[1, 0],
        )
        self._integrated_velocities[1] = np.clip(
            self._integrated_velocities[1],
            0.5 * self._limit_vel_forward_j[0, 1],
            0.5 * self._limit_vel_forward_j[1, 1],
        )
        self.apply_base_velocity(self._integrated_velocities)
        self.apply_velocity_action(self._integrated_velocities)

    def apply_velocity_action_wheels(self, vels: np.ndarray) -> None:
        """Applies angular velocities to the wheels."""
        for axis in range(self._number_actuated_axes):
            for i in range(2):
                p.setJointMotorControl2(
                    self._robot,
                    self._robot_joints[axis * 2 + i],
                    controlMode=p.VELOCITY_CONTROL,
                    targetVelocity=vels[i],
                )

    def apply_base_velocity(self, vels: np.ndarray) -> None:
        """Applies forward and angular velocity to the base.

        The forward and angular velocity of the base
        is first transformed in angular velocities of
        the wheels using a simple dynamics model.

        """
        velocity_left_wheel = (
            vels[0] + 0.5 * self._wheel_distance * vels[1]
        ) / self._wheel_radius
        velocity_right_wheel = (
            vels[0] - 0.5 * self._wheel_distance * vels[1]
        ) / self._wheel_radius

        wheel_velocities = np.array([velocity_left_wheel, velocity_right_wheel])
        self.apply_velocity_action_wheels(wheel_velocities)

    def apply_velocity_action(self, vels: np.ndarray) -> None:
        """Applies angular velocities to the arm joints."""
        self.apply_base_velocity(vels) 
        for i in range(2, self._n):
            p.setJointMotorControl2(
                self._robot,
                self._robot_joints[i],
                controlMode=p.VELOCITY_CONTROL,
                targetVelocity=vels[i],
            )

    def correct_base_orientation(self, pos_base: np.ndarray) -> np.ndarray:
        """Corrects base orientation by -pi.

        The orientation observation should be zero when facing positive
        x-direction. Some robot models are rotated by pi. This is corrected
        here. The function also makes sure that the orientation is always
        between -pi and pi.
        """
        pos_base -= self._spawn_rotation
        if pos_base[2] < -np.pi:
            pos_base[2] += 2 * np.pi
        return pos_base

    def update_state(self) -> None:
        """Updates the robot state.

        The robot state is stored in the self.state, which contains
        a dictionary with key 'joint_state' with nested dictionaries:
        `position`: np.array((base_pose2D, joint_position_2, ...,
            joint_position_n))
            the position in local configuration space
            the base's configuration space aligns with the world frame
            base_pose2D = (x_positions, y_position, orientation)
            the joints 2 to n have al 1-dimensional configuration space
            joint_position_i = (position in local configuration space)
        `velocity`: np.array((base_twist2D, joint_velocity_2, ...,
            joint_velocity_n))
            the velocity in local configuration space
            the base's configuration space aligns with the world frame
            base_pose2D = (x_positions, y_position, orientation)
            the joints 2 to n have al one dimensional configuration space
            joint_velocity_i = (position in local configuration space)
        `forward_velocity`: float
            forward velocity in robot frame
        """
        # base position
        link_state = p.getLinkState(self._robot, 0, computeLinkVelocity=0)
        pos_base = np.array(
            [
                link_state[0][0],
                link_state[0][1],
                p.getEulerFromQuaternion(link_state[1])[2],
            ]
        )
        # make sure that the rotation is within -pi and pi
        self.correct_base_orientation(pos_base)
        # wheel velocities
        vel_wheels = p.getJointStates(self._robot, self._robot_joints)
        v_right = vel_wheels[0][1]
        v_left = vel_wheels[1][1]
        # simple dynamics model to compute the forward and angular velocity
        forward_velocity = 0.5 * (v_right + v_left) * self._wheel_radius
        angular_velocity = (
            (v_right - v_left) * self._wheel_radius / self._wheel_distance
        )

        if self._facing_direction == 'x':
            jacobi_nonholonomic = np.array(
                [[np.cos(pos_base[2]), 0], [np.sin(pos_base[2]), 0], [0, 1]]
            )
        elif self._facing_direction == '-x':
            jacobi_nonholonomic = np.array(
                [[-np.cos(pos_base[2]), 0], [-np.sin(pos_base[2]), 0], [0, 1]]
            )
        elif self._facing_direction == 'y':
            jacobi_nonholonomic = np.array(
                [[-np.sin(pos_base[2]), 0], [np.cos(pos_base[2]), 0], [0, 1]]
            )
        elif self._facing_direction == '-y':
            jacobi_nonholonomic = np.array(
                [[np.sin(pos_base[2]), 0], [-np.cos(pos_base[2]), 0], [0, 1]]
            )
        else:
            raise FacingDirectionUndefinedError(f"Facing direction {self._facing_direction} undefined. Use 'x', '-x', 'y' or '-y'")
        velocity_base = np.dot(
            jacobi_nonholonomic, np.array([forward_velocity, angular_velocity])
        )

        # joint configurations for holonomic joints
        joint_pos_list = []
        joint_vel_list = []
        for i in range(2, self._n):
            pos, vel, _, _ = p.getJointState(self._robot, self._robot_joints[i])
            joint_pos_list.append(pos)
            joint_vel_list.append(vel)
        joint_pos = np.array(joint_pos_list)
        joint_vel = np.array(joint_vel_list)

        self.state = {
            "joint_state": {
                "position": np.concatenate((pos_base, joint_pos)),
                "velocity": np.concatenate((velocity_base, joint_vel)),
                "forward_velocity": np.array([forward_velocity]),
            }
        }




"""Module for free space decomposition."""
import numpy as np
from typing import Callable, List


class HalfPlane(object):
    _normal_vector: np.ndarray
    _point: np.ndarray
    _constant: float

    def __init__(self, point: np.ndarray, position: np.ndarray):
        self._normal_vector = position - point
        self._point = point
        self._constant = -np.dot(self.normal(), point)

    def point_behind_plane(self, point: np.ndarray):
        return np.dot(self.normal(), point) + self.constant() <= 0

    def point_infront_plane(self, point: np.ndarray):
        return not self.point_behind_plane(point)

    def normal(self) -> np.ndarray:
        return self._normal_vector

    def point(self) -> np.ndarray:
        return self._point

    def constant(self) -> float:
        return self._constant

    def equation_by_variable(self, variable_name: str) -> Callable:
        if variable_name == "y":
            return (
                lambda x: 1
                / self.normal()[1]
                * (-self.constant() - self.normal()[0] * x)
            )
        return (
            lambda y: 1
            / self.normal()[0]
            * (-self.constant() - self.normal()[1] * y)
        )

    def get_points(self) -> np.ndarray:
        my_fun = self.equation_by_variable("y")
        x = np.arange(0, 2) * 10 - 5
        y_values = my_fun(x)
        if np.any(np.isinf(y_values)):
            x = np.array([self.point()[0], self.point()[0]])
            y = np.array([-5, 5])
            return np.array([x, y])
        return np.array([x, y_values])

    def constraint(self):
        return np.concatenate((self.normal(), np.array([self.constant()])))


def point_to_point_distance(point_a: np.ndarray, point_b: np.ndarray) -> float:
    return float(np.linalg.norm(point_a - point_b))


class FreeSpaceDecomposition(object):
    _constraints: List[HalfPlane]
    _number_constraints: int
    _max_radius: float
    _position: np.ndarray

    def __init__(
        self,
        position: np.ndarray,
        number_constraints: int = 10,
        max_radius: float = 1.0,
    ):
        self._number_constraints = number_constraints
        self._max_radius = max_radius
        self._constraints = []
        self._position = position

    def set_position(self, position: np.ndarray):
        self._position = position

    def compute_constraints(
        self,
        points: np.ndarray,
    ):
        self._constraints = []
        dists = np.linalg.norm(points - self._position, axis=1)
        idx = np.argsort(dists)
        points = points[idx]
        points = points[dists[idx] < self._max_radius]
        while (
            points.size > 0
            and len(self._constraints) < self._number_constraints
        ):
            point = points[0]
            new_constraint = HalfPlane(point, self._position)
            self._constraints.append(new_constraint)
            mask = np.apply_along_axis(
                new_constraint.point_infront_plane, 1, points
            )
            points = points[mask]

    def constraints(self):
        return self._constraints

    def asdict(self):
        constraint_dict = {}
        for i in range(self._number_constraints):
            if i < len(self._constraints):
                constraint_dict[f"constraint_{i}"] = self._constraints[
                    i
                ].constraint()
            else:
                point = self._position + np.array([20, 20, 0])
                dummy_halfplane = HalfPlane(self._position, point)
                constraint_dict[
                    f"constraint_{i}"
                ] = dummy_halfplane.constraint()
        return constraint_dict

    def get_points(self):
        plot_points = []
        for constraint in self.constraints():
            plot_points.append(constraint.get_points())
        return plot_points



"""Module for fsd sensor based on lidar."""
import numpy as np
import pybullet as p
import gymnasium as gym

from urdfenvs.sensors.lidar import Lidar
from urdfenvs.sensors.fsd_sensor import FSDSensor


class FreeSpaceDecompositionSensor(FSDSensor, Lidar):
    def __init__(
        self,
        link_name,
        max_radius: float = 1.0,
        number_constraints: int = 10,
        nb_rays: int = 10,
        ray_length: float = 10.0,
        angle_limits: np.ndarray = np.array([-np.pi, np.pi]),
        plotting_interval: int = -1,
        plotting_interval_fsd: int = -1,
        variance: float = 0.0,
        planar_visualization: bool = True,
    ):
        FSDSensor.__init__(
            self,
            max_radius,
            number_constraints=number_constraints,
            plotting_interval_fsd=plotting_interval_fsd,
            variance=variance,
            planar_visualization=planar_visualization,
        )
        Lidar.__init__(
            self,
            link_name,
            nb_rays=nb_rays,
            ray_length=ray_length,
            raw_data=False,
            angle_limits=angle_limits,
            variance=variance,
            plotting_interval=plotting_interval,
        )
        self._name = "FreeSpaceDecompSensor"

    def sense(self, robot, obstacles: dict, goals: dict, t: float):
        lidar_observation = Lidar.sense(
            self, robot, obstacles, goals, t
        ).reshape((self._nb_rays, 2))
        lidar_position = np.array(p.getLinkState(robot, self._link_id)[0])
        relative_positions = np.concatenate(
            (
                np.reshape(lidar_observation, (self._nb_rays, 2)),
                np.zeros((self._nb_rays, 1)),
            ),
            axis=1,
        )
        self._height = lidar_position[2]
        absolute_positions = relative_positions + np.repeat(
            lidar_position[np.newaxis, :], self._nb_rays, axis=0
        )
        return self.compute_fsd(absolute_positions, lidar_position)



import numpy as np
import pybullet as p
import gymnasium as gym
from urdfenvs.sensors.sensor import Sensor
from urdfenvs import __version__


class ObstacleSensor(Sensor):
    """
    the ObstacleSensor class is a sensor sensing the exact position of every
    object. The ObstacleSensor is thus a full information sensor which in the
    real world can never exist. The ObstacleSensor returns a dictionary with
    the position of every object when the sense function is called.

    Attributes
    ----------

    _observation: dict
        For every object the Pose and Twist are stored in the observation.
        Pose contains position in cartesian format (x, y, z)
        and orientation in quaternion format (x, y, z, w)
        Twist contains velocity in cartesian format (x, y, z)
        and angular velocity in cartesian format (x, y, z)

    """

    def __init__(self):
        super().__init__("ObstacleSensor")
        self._observation = np.zeros(self.get_observation_size())

    def get_obserrvation_size(self):
        """Getter for the dimension of the observation space."""
        size = 0
        for _ in range(2, p.getNumBodies()):
            size += (
                14  # add space for position, velocity,
                # orientation and angular velocity
            )
        return size

    def get_observation_space(self, obstacles: dict, goals: dict):
        """
        Create observation space, all observed objects should be inside the
        observation space.
        """
        spaces_dict = gym.spaces.Dict()

        min_os_value = -1000
        max_os_value = 1000

        for obj_id in range(2, p.getNumBodies()):
            spaces_dict[f"obstacle_{obj_id-2}"] = gym.spaces.Dict(
                {
                    "pose": gym.spaces.Dict(
                        {
                            "position": gym.spaces.Box(
                                low=min_os_value,
                                high=max_os_value,
                                shape=(3,),
                                dtype=float,
                            ),
                            "orientation": gym.spaces.Box(
                                low=min_os_value,
                                high=max_os_value,
                                shape=(4,),
                                dtype=float,
                            ),
                        }
                    ),
                    "twist": gym.spaces.Dict(
                        {
                            "linear": gym.spaces.Box(
                                low=min_os_value,
                                high=max_os_value,
                                shape=(3,),
                                dtype=float,
                            ),
                            "angular": gym.spaces.Box(
                                low=min_os_value,
                                high=max_os_value,
                                shape=(3,),
                                dtype=float,
                            ),
                        }
                    ),
                }
            )

        return gym.spaces.Dict({self._name: spaces_dict})

    def sense(self, robot, *args):
        """
        Sense the exact position of all the objects.


        """
        observation = {}

        # assumption: p.getBodyInfo(0), p.getBodyInfo(1) are the robot and
        # ground plane respectively
        for obj_id in range(2, p.getNumBodies()):
            pos = p.getBasePositionAndOrientation(obj_id)
            vel = p.getBaseVelocity(obj_id)

            observation[f"obstacle_{obj_id-2}"] = {
                "pose": {
                    "position": np.array(pos[0]),
                    "orientation": np.array(pos[1]),
                },
                "twist": {
                    "linear": np.array(vel[0]),
                    "angular": np.array(vel[1]),
                },
            }

        return observation



import os
import numpy as np

from urdfenvs.urdf_common.quadrotor import QuadrotorModel


class IrisDrone(QuadrotorModel):
    """IRIS model

    Attributes
    ---------
    _scaling: float
        The size scaling in which the urdf should be spawned.
        This also effects the dynamics of the system.

    """

    def __init__(self, mode: str):
        n = 4  # number of actuated joints
        urdf_file = os.path.join(os.path.dirname(__file__), 'iris.urdf')
        self._scaling: float = 1.0
        self._arm_length    = 0.046
        self._k_thrust      = 5.5716e-06
        self._k_drag        = 1.367e-07
        self._rotor_max_rpm = 2500
        self._rotor_min_rpm = 0
        self._spawn_offset: np.ndarray = np.array([0.0, 0.0, 0.00047494])
        super().__init__(n, urdf_file, mode)

    def set_joint_names(self) -> None:
        """Set joint indices.
        - 1: front right, anticlockwise
        - 2: back left, anticlockwise
        - 3: front left, clockwise
        - 4: back right, clockwise
        Top-down view:
        2      3
          x  x
            x   => front
          x  x
        4      1
        """
        self._robot_joints = [1, 2, 3, 4]

    def set_acceleration_limits(self):
        acc_limit = np.array([5.0, 5.0, 5.0])
        self._limit_acc_j[0, :] = -acc_limit
        self._limit_acc_j[1, :] = acc_limit




import pybullet as p
import gymnasium as gym
import numpy as np
import logging


from urdfenvs.urdf_common.generic_robot import GenericRobot


class QuadrotorModel(GenericRobot):
    """ Quadrotor Model for drones

    Attributes
    ----------

    _propeller_radius : float
        The radius of the actuated propellers
    _arm_length : float
        The length of the arm connecting the propellers to the center of mass
    _k_thrust : float
        The thrust coefficient, N/(rad/s)**2
    _k_drag : float
        The drag coefficient, Nm/(rad/s)**2
    _rotor_max_rpm : float
        The maximum rpm of the motor, rad/s
    _rotor_min_rpm : float
        The minimum rpm of the motor, rad/s

    _spawn_offset : np.ndarray
        The offset by which the initial position must be shifted to align
        observation with that position.

    self._pos : np.ndarray
        The position of the quadrotor in the world frame
    self._quat : np.ndarray
        The orientation of the quadrotor in the world frame
    self._vel : np.ndarray
        The linear velocity of the quadrotor in the world frame
    self._omega : np.ndarray
        The angular velocity of the quadrotor in the world frame
    """

    def __init__(self, n: int, urdf_file: str, mode: str) -> None:
        """Constructor for quadrotor model robot."""
        super().__init__(n, urdf_file)
        self._swawn_offset: np.ndarray = np.array(
            [0.0, 0.0, 0.15])  # TODO: check this value

        self._pos = np.zeros(3,dtype=float)
        self._quat = np.array([0., 0., 0., 1.], dtype=float)
        self._vel = np.zeros(3, dtype=float)
        self._omega = np.zeros(3, dtype=float)
        self._rotor_velocity = np.zeros(4, dtype=float)
        self.state = {"joint_state": {"pose": np.zeros(7, dtype=float), "velocity":
            np.zeros(6, dtype=float), "rotor_velocity": np.zeros(4, dtype=float)}}

    def ns(self) -> int:
        """Returns the number of degrees of freedom.

        This is needed as number of actuated joints `_n` is lower that the
        number of degrees of freedom for quadrotor models.
        
        """
        return 6

    def reset(
            self,
            pos: np.ndarray,
            vel: np.ndarray,
            mount_position: np.ndarray,
            mount_orientation: np.ndarray,) -> None:
        """ Reset simulation and add robot """
        logging.warning(
            "The argument 'mount_position' and 'mount_orientation' are \
ignored for drones."
        )
        if hasattr(self, "_robot"):
            p.removeBody(self._robot)
        base_orientation = pos[3:7]
        spawn_pos = self._spawn_offset + np.array([pos[0], pos[1], pos[2]])
        self._robot = p.loadURDF(
            fileName=self._urdf_file,
            basePosition=spawn_pos,
            baseOrientation=base_orientation,
            flags=p.URDF_USE_SELF_COLLISION_EXCLUDE_PARENT,
            globalScaling=self._scaling,
        )

        self.set_joint_names()
        self.extract_joint_ids()
        self.read_limits()
        for i in range(self.n()):
            p.resetJointState(
                self._robot,
                self._robot_joints[i],
                0,
                targetVelocity=vel[i],
            )
        # set base velocity
        self.update_state()

    def check_state(self, pos: np.ndarray, vel: np.ndarray) -> tuple:
        """Filters state of the robot and returns a valid state."""

        if (
            not isinstance(pos, np.ndarray)
            or not pos.size == self.n() + 3
        ):
            pos = np.zeros(self.n() + 3)
        if not isinstance(vel, np.ndarray) or not vel.size == self.n():
            vel = np.zeros(self.n())
        return pos, vel

    def read_limits(self) -> None:
        self._limit_pos_j = np.zeros((2, self.ns() + 1))
        self._limit_vel_j = np.zeros((2, self.ns()))

        # Position limits (x, y, z)
        self._limit_pos_j[0, 0:3] = np.array([-1000., -1000., 0])
        self._limit_pos_j[1, 0:3] = np.array([1000., 1000., 100.])

        # Quaternion limits
        self._limit_pos_j[0, 3:7] = np.array([-1, -1, -1, -1])
        self._limit_pos_j[1, 3:7] = np.array([1, 1, 1, 1])

        # Velocity limits (x, y, z)
        self._limit_vel_j[0, 0:3] = np.array([-40., -40., -40.])
        self._limit_vel_j[1, 0:3] = np.array([40., 40., 40.])
        
        # body rate limits
        self._limit_vel_j[0, 3:6] = np.array([-10., -10., -10.])
        self._limit_vel_j[1, 3:6] = np.array([10., 10., 10.])
        
        # rotor limits
        self._limit_rotors_j = np.zeros((2, 4))
        self._limit_rotors_j[0, :] = np.ones(4) * self._rotor_min_rpm
        self._limit_rotors_j[1, :] = np.ones(4) * self._rotor_max_rpm

    def get_observation_space(self) -> gym.spaces.Dict:
        """Gets the observation space for the quadrotor model.

        The observation space is represented as a dictionary.
        `pose` denotes the pose of the quadrotor in the world frame.
        `velocity` denotes the velocity of the quadrotor in the world frame.
        `rotor_velocities; denotes the rotor velocities.
        """
        return gym.spaces.Dict(
            {
                "joint_state": gym.spaces.Dict({
                    "pose": gym.spaces.Box(
                        low=self._limit_pos_j[0, :],
                        high=self._limit_pos_j[1, :],
                        dtype=float,
                    ),
                    "velocity": gym.spaces.Box(
                        low=self._limit_vel_j[0, :],
                        high=self._limit_vel_j[1, :],
                        dtype=float,
                    ),
                    "rotor_velocity": gym.spaces.Box(
                        low=self._limit_rotors_j[0,:],
                        high=self._limit_rotors_j[1,:],
                        dtype=float,
                    ),
                }),
            }
        )

    def get_velocity_spaces(self) -> tuple:
        """Get observation space and action space when using velocity
        control."""
        ospace = self.get_observation_space()
        uu = self._limit_rotors_j[1, :]
        ul = self._limit_rotors_j[0, :]
        aspace = gym.spaces.Box(low=ul, high=uu, dtype=float)
        return (ospace, aspace)

    def apply_velocity_action(self, vels: np.ndarray) -> None:
        """Applies the propeller speed action to the quadrotor.
        """
        direction = np.array([1, 1, -1, -1])
        for i in range(4):

            p.setJointMotorControl2(
                self._robot,
                self._robot_joints[i],
                controlMode=p.VELOCITY_CONTROL,
                targetVelocity=vels[i] * direction[i],
            )
        self._rotor_velocity = vels.astype('float32')

        self.apply_thrust(vels)
        # self.apply_drag_effect(vels)  # TODO: check literatures for the drag effect

    def apply_acceleration_action(self, accs: np.ndarray) -> None:
        print("Acceleration control not implemented for quadrotor model.")

    def apply_torque_action(self, torques: np.ndarray) -> None:
        """Applies the torques to the quadrotor model.
        """
        print("Torque action is not available for quadrotor model.")

    def correct_base_orientation(self) -> None:
        pass

    def apply_thrust(self, rate: np.ndarray) -> None:
        """PyBullet implementation of a thrust model
        
        Given the rotor rate, calculate the thrust force and moment. 
        The implementation is following Upenn MEAM 620 project 1..

        Parameters
        ----------
        rate : ndarray
            (4)-shaped array of ints containing the rate values of the 4 motors.        
        """
        thrusts = self._k_thrust * np.square(rate)

        k = self._k_drag / self._k_thrust
        l = self._arm_length
        torque_mat = np.array([[1,  1,  1,  1],
                               [0,  l,  0, -l],
                               [-l,  0,  l,  0],
                               [k, -k,  k, -k]])
        u = torque_mat @ thrusts
        torque = u[1:]
        for i in range(4):
            p.applyExternalForce(self._robot,
                                 self._robot_joints[i],
                                 posObj=[0, 0, 0],
                                 forceObj=[0, 0, thrusts[i]],
                                 flags=p.LINK_FRAME)
        p.applyExternalTorque(self._robot,
                              0,
                              torqueObj=torque,
                              flags=p.LINK_FRAME)

    def apply_drag_effect(self, rate: np.ndarray) -> None:
        """PyBullet implementation of a drag model

        Base on the system identification in (Foster, 2015)

        Parameters
        ----------
        rate : ndarray
            (4)-shaped array of ints containing the rate values of the 4 motors.

        """
        base_rot = np.array(
            p.getMatrixFromQuaternion(self._quat)).reshape(3, 3)
        drag_factors = -1 * self._k_drag * np.sum(rate)
        drag = np.dot(base_rot, drag_factors * np.array(self._vel))
        p.applyExternalForce(self._robot,
                             0,
                             forceObj=drag,
                             posObj=[0, 0, 0],
                             flags=p.LINK_FRAME)

    def update_state(self) -> None:
        """Update the robot state.

        The robot joint_state is stored in the dictionary self.state,
        which contains
        `x`: position, np.array([x, y, z])
        `v`: velocity, np.array([vx, vy, vz])
        `q`: attitude, np.array([qx, qy, qz, qw])
        `w`: angular_velocity, np.array([dr, dp, dy])
        """
        # base position
        link_state = p.getLinkState(self._robot, 0, computeLinkVelocity=1)
        self._pos = np.array(
            [link_state[0][0], link_state[0][1], link_state[0][2]], dtype=float
        )
        self._quat = np.array(link_state[1], dtype=float)
        self._vel = np.array(link_state[6], dtype=float)
        self._omega = np.array(link_state[7], dtype=float)

        self.state['joint_state']['pose'] = np.concatenate((self._pos, self._quat))
        self.state['joint_state']['velocity'] = np.concatenate((self._vel, self._omega))
        self.state['joint_state']['rotor_velocity'] = self._rotor_velocity




"""Module for fsd sensor based on lidar."""
import numpy as np
import pybullet as p
import gymnasium as gym

from urdfenvs.sensors.fsd_sensor import FSDSensor
from urdfenvs.sensors.occupancy_sensor import OccupancySensor
from urdfenvs.urdf_common.helpers import extract_link_id


class FreeSpaceOccupancySensor(FSDSensor, OccupancySensor):
    def __init__(
        self,
        link_name,
        name: str = "FreeSpaceDecompSensor",
        max_radius: float = 1.0,
        number_constraints: int = 10,
        limits: np.ndarray = np.array([[-1, 1], [-1, 1], [-1, 1]]),
        resolution: np.ndarray = np.array([10, 10, 10], dtype=int),
        interval: int = -1,
        variance: float = 0.0,
        plotting_interval: int = -1,
        plotting_interval_fsd: int = -1,
        planar_visualization: bool = False,
    ):
        FSDSensor.__init__(
            self,
            max_radius,
            number_constraints=number_constraints,
            plotting_interval_fsd=plotting_interval_fsd,
            planar_visualization=planar_visualization,
            variance=variance,
        )
        OccupancySensor.__init__(
            self,
            limits=limits,
            resolution=resolution,
            interval=interval,
            variance=variance,
            plotting_interval=plotting_interval,
        )
        self._link_name = link_name
        self._name = name
        self._height = 0.3

    def sense(self, robot, obstacles: dict, goals: dict, t: float):
        occupancy = OccupancySensor.sense(
            self, robot, obstacles, goals, t
        ).reshape((-1, 1))
        link_id = extract_link_id(robot, self._link_name)
        seed_point = np.array(p.getLinkState(robot, link_id)[0])
        point_positions = self._mesh_flat[np.where(occupancy == 1)[0]]
        return self.compute_fsd(point_positions, seed_point)



"""Module for occupancy sensor simulation."""
from abc import abstractmethod
from typing import Tuple

import numpy as np

from urdfenvs.sensors.sensor import Sensor


class LinkIdNotFoundError(Exception):
    pass


class GridSensor(Sensor):
    def __init__(
        self,
        limits: np.ndarray = np.array([[-1, -1], [-1, 1], [-1, 1]]),
        resolution: np.ndarray = np.array([10, 10, 10], dtype=int),
        interval: int = -1,
        name: str = "Grid",
        variance: float = 0.0,
        plotting_interval: int = -1,
    ):
        super().__init__(
            name, variance=variance, plotting_interval=plotting_interval
        )
        self._resolution = resolution
        self._limits = limits
        self._interval = interval
        self._compute_call_counter = 13
        self._computed = False
        x_values = np.linspace(limits[0][0], limits[0][1], resolution[0])
        y_values = np.linspace(limits[1][0], limits[1][1], resolution[1])
        z_values = np.linspace(limits[2][0], limits[2][1], resolution[2])
        self._mesh = np.stack(
            np.meshgrid(x_values, y_values, z_values, indexing="ij"), axis=-1
        )
        self._mesh_flat = self.mesh().reshape((-1, 3))
        self._grid_values = np.zeros(shape=self._mesh.shape[0:3], dtype=int)

    def mesh(self) -> np.ndarray:
        return self._mesh

    def number_of_voxels(self) -> int:
        return self._mesh.size // 3

    def voxel_size(self) -> Tuple[float]:
        voxel_size = (self._limits[:, 1] - self._limits[:, 0]) / (
            self._resolution - 1
        )
        nan_index = np.where(np.isnan(voxel_size))[0]
        voxel_size[nan_index] = np.maximum(
            0.01, self._limits[nan_index, 1] - self._limits[nan_index, 0]
        )
        return voxel_size

    def get_observation_size(self):
        return self._grid_values.shape

    @abstractmethod
    def get_observation_space(self, obstacles: dict, goals: dict):
        """Create observation space, all observations should be inside the
        observation space."""
        pass

    def distances(self, obstacles: dict, t: float) -> np.ndarray:
        max_voxel_size = np.max(self.voxel_size()) / 2.0
        mesh_flat = self._mesh.reshape((-1, 3))
        if obstacles:
            distances = np.min(
                np.array(
                    [
                        obstacle.distance(mesh_flat, t=t) - max_voxel_size
                        for obstacle in list(obstacles.values())
                    ]
                ),
                axis=0,
            )
            noisy_distances = np.random.normal(distances, self._variance)
            return noisy_distances
        else:
            return np.ones(self.number_of_voxels()) * 1000

    @abstractmethod
    def sense(
        self, robot, obstacles: dict, goals: dict, t: float
    ) -> np.ndarray:
        pass
